# Copyright (c) Dietmar Wolz.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory.
from __future__ import annotations

""" Numpy based implementation of an diversifying wrapper / parallel retry mechanism. 

Uses the archive from CVT MAP-Elites (https://arxiv.org/abs/1610.05729)
and generalizes ideas from CMA-ME (https://arxiv.org/pdf/1912.02400.pdf)
to other wrapped algorithms. 

Both the parallel retry and the archive based modification of the fitness 
function enhance the diversification of the optimization result.
The resulting archive may be stored and can be used to continue the
optimization later.   

Requires a QD-fitness function returning both an fitness value and a
behavior vector used to determine the corresponding archive niche using
Voronoi tesselation. 

Returns an archive of niche-elites containing also for each niche statistics 
about the associated solutions.     
"""

import numpy as np
from numpy.random import Generator, MT19937, SeedSequence
from multiprocessing import Process
from scipy.optimize import Bounds
from fcmaes.optimizer import dtime, de_cma, Optimizer
import multiprocessing as mp
import ctypes as ct
from time import perf_counter
from fcmaes.mapelites import Archive, update_archive, rng
from fcmaes import advretry
from fcmaes.evaluator import is_debug_active
from loguru import logger
import threadpoolctl

from typing import Optional, Callable, Tuple, Dict
from numpy.typing import ArrayLike

def minimize(qd_fitness: Callable[[ArrayLike], Tuple[float, np.ndarray]], 
            bounds: Bounds,
            qd_bounds: Bounds,
            niche_num: Optional[int] = 10000,
            samples_per_niche: Optional[int] = 20,
            max_evals: Optional[int] = None,
            workers: Optional[int] = mp.cpu_count(),
            archive: Optional[Archive] = None,
            opt_params: Optional[Dict] = {},
            use_stats: Optional[bool] = False,            
            ) -> Archive:
    
    """Wraps an fcmaes optmizer/solver by hijacking its tell function.
    Works as CVT Map-Elites in maintaining an archive of diverse elites. 
    But this archive is not used to derive solution vectors, but to reevaluate them. 
    For each fitness result it determines its niche. The "told" fitness is
    determined relative to its local elite. If it is better the evaluated solution
    becomes the new niche-elite.  
    This way the wrapped solver is "tricked" to follow a QD-goal: Finding empty niches
    and improving all niches. This works not only for CMA-ES, but also for other 
    solvers: DE, CR-FM-NES and PGPE. Both their Python and C++ versions are supported. 
     
    Parameters
    ----------
    solver : evolutionary algorithm, needs to support ask/tell 
    qd_fitness : callable
        The objective function to be minimized. Returns a fitness value and a behavior vector. 
            ``qd_fitness(x) -> float, array``
        where ``x`` is an 1-D array with shape (n,)
    bounds : `Bounds`
        Bounds on variables. Instance of the `scipy.Bounds` class.
    qd_bounds : `Bounds`
        Bounds on behavior descriptors. Instance of the `scipy.Bounds` class.        
    niche_num : int, optional
        Number of niches.
    samples_per_niche : int, optional
        Number of samples used for niche computation. 
        If samples_per_niche > 0 cvt-clustering is used, else grid-clustering is used. 
    max_evals : int, optional
        Number of fitness evaluations.
    workers : int, optional
        Number of spawned parallel worker processes.
    archive : Archive, optional
        If defined MAP-elites is continued for this archive.
    opt_params : dictionary, optional (or a list/tuple/array of these)
        Parameters selecting and configuring the wrapped solver.
        'solver' - supported are 'CMA','CMA_CPP','CRMFNES','CRMFNES_CPP','DE','DE_CPP','PGPE'
                    default is 'CMA_CPP'
        'popsize' - population size, default = 32
        'sigma' -  initial distribution sigma, default = rg.uniform(0.03, 0.3)**2)
        'mean' - initial distribution mean, default=rg.uniform(bounds.lb, bounds.ub)) 
        'max_evals' - maximal number of evaluations per run, default = 50000
        'stall_criterion' - how many iterations without progress allowed, default = 50 iterations 
        If a list/tuple/array of parameters are given, the corresponding solvers are called in a 
        sequence.      
    use_stats : bool, optional 
        If True, archive accumulates statistics of the solutions
                
    Returns
    -------
    archive : Archive
        Resulting archive of niches. Can be stored for later continuation of MAP-elites."""

    if max_evals is None:
        max_evals = workers*50000
    dim = len(bounds.lb)
    if archive is None: 
        archive = Archive(dim, qd_bounds, niche_num, use_stats)
        archive.init_niches(samples_per_niche)
        # initialize archive with random values
        archive.xs_view[:] = rng.uniform(bounds.lb, bounds.ub, (niche_num, dim))       
    t0 = perf_counter()   
    qd_fitness.archive = archive # attach archive for logging     
    minimize_parallel_(archive, qd_fitness, bounds, workers, opt_params, max_evals)
    if is_debug_active():
        ys = np.sort(archive.get_ys())[:min(100, archive.capacity)] # best fitness values
        logger.debug(f'best {min(ys):.3f} worst {max(ys):.3f} ' + 
                 f'mean {np.mean(ys):.3f} stdev {np.std(ys):.3f} time {dtime(t0)} s')
    return archive

def apply_advretry(fitness: Callable[[ArrayLike], float], 
                   qd_fitness: Callable[[ArrayLike], Tuple[float, np.ndarray]], 
                   bounds: Bounds, 
                   archive: Archive, 
                   optimizer: Optional[Optimizer] = None, 
                   num_retries: Optional[int] = 1000, 
                   workers: Optional[int] = mp.cpu_count(),
                   max_eval_fac: Optional[float] = 5.0,
                   xs: Optional[np.ndarray] = None,
                   ys: Optional[np.ndarray] = None,
                   x_conv: Callable[[ArrayLike], ArrayLike] = None):
        
    """Unifies the QD world with traditional optimization. It converts
    a QD-archive into a multiprocessing store used by the fcmaes smart
    boundary management meta algorithm (advretry). Then advretry is applied
    to find the global optimum. Finally the updated store is feed back into
    the QD-archive. For this we need a descriptor generating function 
    'descriptors' which may require reevaluation of the new solutions.  
        
    Parameters
    ----------
    solver : evolutionary algorithm, needs to support ask/tell 
    fitness : callable
        The objective function to be minimized. Returns a fitness value. 
            ``fitness(x) -> float``
    qf_fun : callable
        Generates the descriptors for a solution. Returns a behavior vector. 
            ``descriptors(x) -> array``
        where ``x`` is an 1-D array with shape (n,)
    bounds : `Bounds`
        Bounds on variables. Instance of the `scipy.Bounds` class.
    archive : Archive
        Improves the solutions if this archive.
    optimizer : optimizer.Optimizer, optional
        Optimizer to use. Default is a sequence of differential evolution and CMA-ES.
    num_retries : int, optional
        Number of optimization runs.
    workers : int, optional
        Number of spawned parallel worker processes.
    max_eval_fac : int, optional
        Final limit of the number of function evaluations = max_eval_fac*min_evaluations  
    xs : ndarray, optional
        Used to initialize advretry. If undefined the archive content is used. 
        If xs is defined, ys must be too
    ys : ndarray, optional
        Used to initialize advretry. If undefined the archive content is used.  
    x_conv : callable, optional
        If defined converts the x in xs to solutions suitable for the given archive.
        If undefined it is assumed that the x in xs are valid archive solutons.   
    """

    if optimizer is None:
        optimizer = de_cma(1500)
    # generate advretry store
    store = advretry.Store(fitness, bounds, num_retries=num_retries, 
                           max_eval_fac=max_eval_fac)  
                         
    # select only occupied entries
    if xs is None:
        ys = archive.get_ys()    
        valid = (ys < np.inf)
        ys = ys[valid]
        xs = archive.xs_view[valid]
    t0 = perf_counter() 
    # transfer to advretry store
    for i in range(len(ys)):
        store.add_result(ys[i], xs[i], 1)
    # perform parallel retry
    advretry.retry(store, optimizer.minimize, workers=workers)
    # transfer back to archive 
    xs = store.xs_view
    if not x_conv is None:
        xs = [x_conv(x) for x in xs]
    yds = [qd_fitness(x) for x in xs]
    descs = np.array([yd[1] for yd in yds])
    ys = np.array([yd[0] for yd in yds])
    niches = archive.index_of_niches(descs)
    for i in range(len(ys)):
        archive.set(niches[i], (ys[i], descs[i]), xs[i])
    archive.argsort()
    if is_debug_active():
        ys = np.sort(archive.get_ys())[:min(100, archive.capacity)] # best fitness values
        logger.debug(f'best {min(ys):.3f} worst {max(ys):.3f} ' + 
                 f'mean {np.mean(ys):.3f} stdev {np.std(ys):.3f} time {dtime(t0)} s')    

def minimize_parallel_(archive, fitness, bounds, workers, opt_params, max_evals):
    sg = SeedSequence()
    rgs = [Generator(MT19937(s)) for s in sg.spawn(workers)]
    evals = mp.RawValue(ct.c_long, 0)
    proc=[Process(target=run_minimize_,
            args=(archive, fitness, bounds, rgs[p],
                  opt_params, p, workers, evals, max_evals)) for p in range(workers)]
    [p.start() for p in proc]
    [p.join() for p in proc]
                    
def run_minimize_(archive, fitness, bounds, rg, opt_params, p, workers, evals, max_evals):  
    with threadpoolctl.threadpool_limits(limits=1, user_api="blas"):
        if isinstance(opt_params, (list, tuple, np.ndarray)):
            default_workers = int(workers/2) if len(opt_params) > 1 else workers
            for params in opt_params: # call MAP-Elites
                if 'elites' == params.get('solver'):
                    elites_workers = params.get('workers', default_workers) 
                    if p < elites_workers:
                        run_map_elites_(archive, fitness, bounds, rg, evals, max_evals, params)
                        return
        while evals.value < max_evals: # call solvers in loop     
            best_x = None
            if isinstance(opt_params, (list, tuple, np.ndarray)):
                for params in opt_params: # call in sequence
                    if 'elites' == params.get('solver'):
                        continue # ignore in loop
                    if best_x is None:
                        # selecting a niche elite is no improvement over random x0
                        x0 = None#, _, _ = archive.random_xs_one(select_n, rg)
                        best_x = minimize_(archive, fitness, bounds, rg, evals, max_evals, params, 
                                           x0 = x0)
                    else:
                        best_x = minimize_(archive, fitness, bounds, rg, evals, max_evals, params, x0 = best_x)
            else:        
                minimize_(archive, fitness, bounds, rg, evals, max_evals, opt_params) 

from fcmaes.mapelites import variation_,  iso_dd_
                
def run_map_elites_(archive, fitness, bounds, rg, evals, max_evals, opt_params = {}):    
    popsize = opt_params.get('popsize', 32)  
    use_sbx = opt_params.get('use_sbx', True)     
    dis_c = opt_params.get('dis_c', 20)   
    dis_m = opt_params.get('dis_m', 20)  
    iso_sigma = opt_params.get('iso_sigma', 0.01)
    line_sigma = opt_params.get('line_sigma', 0.2)
    select_n = archive.capacity
    while evals.value < max_evals:              
        if use_sbx:
            pop = archive.random_xs(select_n, popsize, rg)
            xs = variation_(pop, bounds.lb, bounds.ub, rg, dis_c, dis_m)
        else:
            x1 = archive.random_xs(select_n, popsize, rg)
            x2 = archive.random_xs(select_n, popsize, rg)
            xs = iso_dd_(x1, x2, bounds.lb, bounds.ub, rg, iso_sigma, line_sigma)    
        yds = [fitness(x) for x in xs]
        evals.value += popsize
        descs = np.array([yd[1] for yd in yds])
        niches = archive.index_of_niches(descs)
        for i in range(len(yds)):
            archive.set(niches[i], yds[i], xs[i])
        archive.argsort()   
        select_n = archive.get_occupied()  

def minimize_(archive, fitness, bounds, rg, evals, max_evals, opt_params, x0 = None): 
    if 'BITE_CPP' == opt_params.get('solver'):
        return run_bite_(archive, fitness, bounds, rg, evals, max_evals, opt_params, x0 = None)
    else:
        es = get_solver_(bounds, opt_params, rg, x0) 
        stall_criterion = opt_params.get('stall_criterion', 20)
        max_evals_iter = opt_params.get('max_evals', 50000)
        max_iters = int(max_evals_iter/es.popsize)
        old_ys = None
        last_improve = 0
        best_x = None
        best_y = np.inf
        for iter in range(max_iters):
            xs = es.ask()
            ys, real_ys = update_archive(archive, xs, fitness)
            evals.value += es.popsize
            # update best real fitness
            yi = np.argmin(real_ys)
            ybest = real_ys[yi] 
            if ybest < best_y:
                best_y = ybest
                best_x = xs[yi]
            if not old_ys is None:
                if (np.sort(ys) < old_ys).any():
                    last_improve = iter          
            if last_improve + stall_criterion < iter:
                break
            stop = es.tell(ys)
            if stop != 0 or evals.value >= max_evals:
                break 
            old_ys = np.sort(ys)
        return best_x # real best solution

from fcmaes import cmaes, cmaescpp, crfmnescpp, pgpecpp, decpp, crfmnes, de, bitecpp

def run_bite_(archive, fitness, bounds, rg, evals, max_evals, opt_params, x0 = None):  
    # BiteOpt doesn't support ask/tell, so we have to "patch" fitness. Note that Voronoi 
    # tesselation is more expensive if called for single behavior vectors and not for batches. 
    
    def fit(x: Callable[[ArrayLike], float]):
        if evals.value >= max_evals:
            return np.inf
        evals.value += 1
        ys, _ = update_archive(archive, [x], fitness)
        return ys[0]
    
    max_evals_iter = opt_params.get('max_evals', 50000)       
    stall_criterion = opt_params.get('stall_criterion', 20)   
    #popsize = opt_params.get('popsize', 0) 
    ret = bitecpp.minimize(fit, bounds, x0 = x0, M = 1, 
                           stall_criterion = stall_criterion,
                           max_evaluations = max_evals_iter, rg = rg)
    return ret.x   

def get_solver_(bounds, opt_params, rg, x0 = None):
    dim = len(bounds.lb)
    popsize = opt_params.get('popsize', 31) 
    #sigma = opt_params.get('sigma',rg.uniform(0.03, 0.3)**2)
    sigma = opt_params.get('sigma',rg.uniform(0.1, 0.5)**2)
    #sigma = opt_params.get('sigma',rg.uniform(0.2, 0.5)**2)
    #sigma = opt_params.get('sigma',rg.uniform(0.1, 0.5))
    mean = opt_params.get('mean', rg.uniform(bounds.lb, bounds.ub)) \
                if x0 is None else x0
    name = opt_params.get('solver', 'CMA_CPP')
    if name == 'CMA':
        return cmaes.Cmaes(bounds, x0 = mean,
                          popsize = popsize, input_sigma = sigma, rg = rg)
    elif name == 'CMA_CPP':
        return cmaescpp.ACMA_C(dim, bounds, x0 = mean, #stop_hist = 0,
                          popsize = popsize, input_sigma = sigma, rg = rg)
    elif name == 'CRMFNES':
        return crfmnes.CRFMNES(dim, bounds, x0 = mean,
                          popsize = popsize, input_sigma = sigma, rg = rg)
    elif name == 'CRMFNES_CPP':
        return crfmnescpp.CRFMNES_C(dim, bounds, x0 = mean,
                          popsize = popsize, input_sigma = sigma, rg = rg)
    elif name == 'DE':
        return de.DE(dim, bounds, popsize = popsize, rg = rg)
    elif name == 'DE_CPP':
        return decpp.DE_C(dim, bounds, popsize = popsize, rg = rg)
    elif name == 'PGPE':
        return pgpecpp.PGPE_C(dim, bounds, x0 = mean,
                          popsize = popsize, input_sigma = sigma, rg = rg)
    else:
        print ("invalid solver")
        return None
            



# Copyright (c) Dietmar Wolz.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory.

"""
Eigen based implementation of PGPE see http://mediatum.ub.tum.de/doc/1099128/631352.pdf .
Derived from https://github.com/google/evojax/blob/main/evojax/algo/pgpe.py .
"""

import sys
import os
import math
import ctypes as ct
import numpy as np
from numpy.random import MT19937, Generator
from scipy.optimize import OptimizeResult, Bounds
from fcmaes.evaluator import _check_bounds, _get_bounds, callback_par, parallel, call_back_par, libcmalib

from typing import Optional, Callable, Union
from numpy.typing import ArrayLike

os.environ['MKL_DEBUG_CPU_TYPE'] = '5'

def minimize(fun: Callable[[ArrayLike], float],
             bounds: Optional[Bounds] = None,
             x0: Optional[ArrayLike] = None,
             input_sigma: Optional[Union[float, ArrayLike, Callable]] = 0.1,
             popsize: Optional[int] = 32,
             max_evaluations: Optional[int] = 100000,
             workers: Optional[int] = None,
             stop_fitness: Optional[float] = -np.inf,
             rg: Optional[Generator] = Generator(MT19937()),
             runid: Optional[int] = 0,
             normalize: Optional[bool] = True,
             lr_decay_steps: Optional[int] = 1000,
             use_ranking: Optional[bool] = True,
             center_learning_rate: Optional[float] = 0.15,
             stdev_learning_rate: Optional[float] = 0.1,
             stdev_max_change: Optional[float] = 0.2,
             b1: Optional[float] = 0.9,
             b2: Optional[float] = 0.999,
             eps: Optional[float] = 1e-8,
             decay_coef: Optional[float] = 1.0,
             ) -> OptimizeResult:
       
    """Minimization of a scalar function of one or more variables using a 
    C++ PGPE implementation called via ctypes.
     
    Parameters
    ----------
    fun : callable
        The objective function to be minimized.
            ``fun(x) -> float``
        where ``x`` is an 1-D array with shape (dim,)
    bounds : sequence or `Bounds`, optional
        Bounds on variables. There are two ways to specify the bounds:
            1. Instance of the `scipy.Bounds` class.
            2. Sequence of ``(min, max)`` pairs for each element in `x`. None
               is used to specify no bound.
    x0 : ndarray, shape (dim,)
        Initial guess. Array of real elements of size (dim,),
        where 'dim' is the number of independent variables.  
    input_sigma : float or np.array, optional
        Initial step size.
    popsize = int, optional
        CMA-ES population size.
    max_evaluations : int, optional
        Forced termination after ``max_evaluations`` function evaluations.
    workers : int or None, optional
        If workers > 1, function evaluation is performed in parallel for the whole population. 
        Useful for costly objective functions but is deactivated for parallel retry.  
    stop_fitness : float, optional 
         Limit for fitness value. If reached minimize terminates.
    rg = numpy.random.Generator, optional
        Random generator for creating random guesses.
    runid : int, optional
        id used by the is_terminate callback to identify the CMA-ES run.     
    normalize : boolean, optional
        if true pheno -> geno transformation maps arguments to interval [-1,1] 
    lr_decay_steps int, optional - ADAM optimizer configuration
    use_ranking : boolean, optional - Should we treat the fitness as rankings or not.       
    center_learning_rate : float, optional - Learning rate for the Gaussian mean.
    stdev_learning_rate : float, optional - Learning rate for the Gaussian stdev.
    init_stdev : float, optional - Initial stdev for the Gaussian distribution.
    stdev_max_change : float, optional - Maximum allowed change for stdev in abs values.    
    b1 : float, optional - ADAM optimizer configuration
    b2 : float, optional - ADAM optimizer configuration
    eps : float, optional - ADAM optimizer configuration
    decay_coef : float, optional - ADAM optimizer configuration
        
    Returns
    -------
    res : scipy.OptimizeResult
        The optimization result is represented as an ``OptimizeResult`` object.
        Important attributes are: ``x`` the solution array, 
        ``fun`` the best function value, 
        ``nfev`` the number of function evaluations,
        ``nit`` the number of CMA-ES iterations, 
        ``status`` the stopping critera and
        ``success`` a Boolean flag indicating if the optimizer exited successfully. """
    
    lower, upper, guess = _check_bounds(bounds, x0, rg)      
    dim = guess.size   
    if popsize is None:
        popsize = 32      
    if popsize % 2 == 1: # requires even popsize
        popsize += 1
    if lower is None:
        lower = [0]*dim
        upper = [0]*dim
    if callable(input_sigma):
        input_sigma=input_sigma()
    if np.ndim(input_sigma) == 0:
        input_sigma = [input_sigma] * dim 
    parfun = None if (workers is None or workers <= 1) else parallel(fun, workers)
    array_type = ct.c_double * dim   
    c_callback_par = call_back_par(callback_par(fun, parfun))
    res = np.empty(dim+4)
    res_p = res.ctypes.data_as(ct.POINTER(ct.c_double))
    try:
        optimizePGPE_C(runid, c_callback_par, dim, array_type(*guess), 
                array_type(*lower), array_type(*upper), 
                array_type(*input_sigma), max_evaluations, stop_fitness,
                popsize, int(rg.uniform(0, 2**32 - 1)), 
                lr_decay_steps, use_ranking, center_learning_rate,
                stdev_learning_rate, stdev_max_change, b1, b2, eps, decay_coef,
                normalize, res_p)
        x = res[:dim]
        val = res[dim]
        evals = int(res[dim+1])
        iterations = int(res[dim+2])
        stop = int(res[dim+3])
        res = OptimizeResult(x=x, fun=val, nfev=evals, nit=iterations, status=stop, success=True)
    except Exception as ex:
        res = OptimizeResult(x=None, fun=sys.float_info.max, nfev=0, nit=0, status=-1, success=False)
    if not parfun is None:
        parfun.stop()
    return res

class PGPE_C:

    def __init__(self,
        dim: int,
        bounds: Optional[Bounds] = None,
        x0: Optional[ArrayLike] = None,
        input_sigma: Optional[Union[float, ArrayLike, Callable]] = 0.1,
        popsize: Optional[int] = 32,
        rg: Optional[Generator] = Generator(MT19937()),
        runid: Optional[int] = 0,
        normalize: Optional[bool] = True,
        lr_decay_steps: Optional[int] = 1000,
        use_ranking: Optional[bool] = False, 
        center_learning_rate: Optional[float] = 0.15,
        stdev_learning_rate: Optional[float] = 0.1, 
        stdev_max_change: Optional[float] = 0.2, 
        b1: Optional[float] = 0.9,
        b2: Optional[float] = 0.999, 
        eps: Optional[float] = 1e-8, 
        decay_coef: Optional[float] = 1.0, 
        ):
       
        """Minimization of a scalar function of one or more variables using a 
        C++ CR-FM-NES implementation called via ctypes.
         
        Parameters
        ----------
        dim : int
            dimension of the argument of the objective function
        bounds : sequence or `Bounds`, optional
            Bounds on variables. There are two ways to specify the bounds:
                1. Instance of the `scipy.Bounds` class.
                2. Sequence of ``(min, max)`` pairs for each element in `x`. None
                   is used to specify no bound.
        x0 : ndarray, shape (dim,)
            Initial guess. Array of real elements of size (dim,),
            where 'dim' is the number of independent variables.  
        input_sigma : float, optional
            Initial step size.
        popsize = int, optional
            CMA-ES population size.
        max_evaluations : int, optional
            Forced termination after ``max_evaluations`` function evaluations.
        workers : int or None, optional
            If workers > 1, function evaluation is performed in parallel for the whole population. 
            Useful for costly objective functions but is deactivated for parallel retry.  
        stop_fitness : float, optional 
             Limit for fitness value. If reached minimize terminates.
        rg = numpy.random.Generator, optional
            Random generator for creating random guesses.
        runid : int, optional
            id used by the is_terminate callback to identify the CMA-ES run.     
        normalize : boolean, optional
            if true pheno -> geno transformation maps arguments to interval [-1,1]
        lr_decay_steps int, optional - ADAM optimizer configuration
        use_ranking : boolean, optional - Should we treat the fitness as rankings or not.       
        center_learning_rate : float, optional - Learning rate for the Gaussian mean.
        stdev_learning_rate : float, optional - Learning rate for the Gaussian stdev.
        init_stdev : float, optional - Initial stdev for the Gaussian distribution.
        stdev_max_change : float, optional - Maximum allowed change for stdev in abs values.    
        b1 : float, optional - ADAM optimizer configuration
        b2 : float, optional - ADAM optimizer configuration
        eps : float, optional - ADAM optimizer configuration
        decay_coef : float, optional - ADAM optimizer configuration"""

        lower, upper, guess = _get_bounds(dim, bounds, x0, rg)      
        if popsize is None:
            popsize = 32      
        if popsize % 2 == 1: # requires even popsize
            popsize += 1
        if lower is None:
            lower = [0]*dim
            upper = [0]*dim
        if callable(input_sigma):
            input_sigma=input_sigma()
        if np.ndim(input_sigma) == 0:
            input_sigma = [input_sigma] * dim 
        array_type = ct.c_double * dim   
        try:
            self.ptr = initPGPE_C(runid, dim, array_type(*guess), 
                           array_type(*lower), array_type(*upper), 
                    array_type(*input_sigma), popsize, int(rg.uniform(0, 2**32 - 1)),
                    lr_decay_steps, use_ranking, center_learning_rate,
                    stdev_learning_rate, stdev_max_change, b1, b2, eps, decay_coef, 
                    normalize)
            self.popsize = popsize
            self.dim = dim            
        except Exception as ex:
            print (ex)
            pass
    
    def __del__(self):
        destroyPGPE_C(self.ptr)
            
    def ask(self) -> np.array:
        """ask for popsize new argument vectors.
            
        Returns
        -------
        xs : popsize sized list of dim sized argument lists."""
        
        try:
            lamb = self.popsize
            n = self.dim
            res = np.empty(lamb*n)
            res_p = res.ctypes.data_as(ct.POINTER(ct.c_double))
            askPGPE_C(self.ptr, res_p)
            xs = np.empty((lamb, n))
            for p in range(lamb):
                xs[p,:] = res[p*n : (p+1)*n]
            return xs
        except Exception as ex:
            print (ex)
            return None

    def tell(self, 
             ys: np.ndarray) -> int:      
        """tell function values for the argument lists retrieved by ask().
        
        Parameters
        ----------
        ys : popsize sized list of function values 
 
        Returns
        -------
        stop : int termination criteria, if != 0 loop should stop."""        
        
        try:
            array_type_ys = ct.c_double * len(ys)
            return tellPGPE_C(self.ptr, array_type_ys(*ys))
        except Exception as ex:
            print (ex)
            return -1        

    def population(self) -> np.array:
        try:
            lamb = self.popsize
            n = self.dim
            res = np.empty(lamb*n)
            res_p = res.ctypes.data_as(ct.POINTER(ct.c_double))
            populationPGPE_C(self.ptr, res_p)
            xs = np.array(lamb, n)
            for p in range(lamb):
                xs[p] = res[p*n : (p+1)*n]
                return xs
        except Exception as ex:
            print (ex)
            return None

    def result(self) -> OptimizeResult:
        res = np.empty(self.dim+4)
        res_p = res.ctypes.data_as(ct.POINTER(ct.c_double))
        try:
            resultPGPE_C(self.ptr, res_p)
            x = res[:self.dim]
            val = res[self.dim]
            evals = int(res[self.dim+1])
            iterations = int(res[self.dim+2])
            stop = int(res[self.dim+3])
            res = OptimizeResult(x=x, fun=val, nfev=evals, nit=iterations, status=stop, success=True)
        except Exception as ex:
            res = OptimizeResult(x=None, fun=sys.float_info.max, nfev=0, nit=0, status=-1, success=False)
        return res

if not libcmalib is None: 

    optimizePGPE_C = libcmalib.optimizePGPE_C
    optimizePGPE_C.argtypes = [ct.c_long, call_back_par, ct.c_int, \
                ct.POINTER(ct.c_double), ct.POINTER(ct.c_double), ct.POINTER(ct.c_double), \
                ct.POINTER(ct.c_double), ct.c_int, ct.c_double, ct.c_int, 
                ct.c_long, ct.c_int, 
                ct.c_bool, ct.c_double, ct.c_double, ct.c_double, 
                ct.c_double, ct.c_double, ct.c_double, ct.c_double, 
                ct.c_bool, ct.POINTER(ct.c_double)]
                
    initPGPE_C = libcmalib.initPGPE_C
    initPGPE_C.argtypes = [ct.c_long, ct.c_int, ct.POINTER(ct.c_double), 
                ct.POINTER(ct.c_double), ct.POINTER(ct.c_double), ct.POINTER(ct.c_double), 
                ct.c_int, ct.c_long, ct.c_int, 
                ct.c_bool, ct.c_double, ct.c_double, ct.c_double, 
                ct.c_double, ct.c_double, ct.c_double, ct.c_double, 
                ct.c_bool]
    
    initPGPE_C.restype = ct.c_void_p   
    
    destroyPGPE_C = libcmalib.destroyPGPE_C
    destroyPGPE_C.argtypes = [ct.c_void_p]
    
    askPGPE_C = libcmalib.askPGPE_C
    askPGPE_C.argtypes = [ct.c_void_p, ct.POINTER(ct.c_double)]
    
    tellPGPE_C = libcmalib.tellPGPE_C
    tellPGPE_C.argtypes = [ct.c_void_p, ct.POINTER(ct.c_double)]
    tellPGPE_C.restype = ct.c_int
    
    populationPGPE_C = libcmalib.populationPGPE_C
    populationPGPE_C.argtypes = [ct.c_void_p, ct.POINTER(ct.c_double)]
    
    resultPGPE_C = libcmalib.resultPGPE_C
    resultPGPE_C.argtypes = [ct.c_void_p, ct.POINTER(ct.c_double)]



# Copyright (c) Dietmar Wolz.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory.

import numpy as np
from numpy.random import MT19937, Generator
from scipy.optimize import Bounds, minimize, shgo, differential_evolution, dual_annealing, basinhopping
import sys
import time
from loguru import logger
import ctypes as ct
import multiprocessing as mp 
from fcmaes.evaluator import serial, parallel
from fcmaes import crfmnes, crfmnescpp, pgpecpp, cmaes, de, cmaescpp, decpp, dacpp, gcldecpp, lcldecpp, ldecpp, csmacpp, bitecpp

from typing import Optional, Callable, Tuple, Union
from numpy.typing import ArrayLike

def eprint(*args, **kwargs):
    """print message to stderr."""
    print(*args, file=sys.stderr, **kwargs)

def scale(lower: ArrayLike, 
          upper: ArrayLike) -> np.ndarray:
    """scaling = 0.5 * difference of the bounds."""
    return 0.5 * (np.asarray(upper) - np.asarray(lower))

def typical(lower: ArrayLike, 
            upper: ArrayLike) -> np.ndarray:
    """typical value = mean of the bounds."""
    return 0.5 * (np.asarray(upper) + np.asarray(lower))

def fitting(guess: ArrayLike, 
            lower: ArrayLike, 
            upper: ArrayLike) -> np.ndarray:
    """fit a guess into the bounds."""
    return np.clip(np.asarray(guess), np.asarray(upper), np.asarray(lower))

def is_terminate(runid: int, 
                 iterations: int, 
                 val: float) -> bool:
    """dummy is_terminate call back."""
    return False    

def random_x(lower: ArrayLike, upper: ArrayLike) -> np.ndarray:
    """feasible random value uniformly distributed inside the bounds."""
    lower = np.asarray(lower)
    upper = np.asarray(upper)
    return lower + np.multiply(upper - lower, np.random.rand(lower.size))
    
def dtime(t0: float) -> float:
    """time since t0."""
    return round(time.perf_counter() - t0, 2)

class wrapper(object):
    """Fitness function wrapper for use with parallel retry."""

    def __init__(self, 
                 fit: Callable[[ArrayLike], float]):
        self.fit = fit
        self.evals = mp.RawValue(ct.c_int, 0) 
        self.best_y = mp.RawValue(ct.c_double, np.inf) 
        self.t0 = time.perf_counter()

    def __call__(self, x: ArrayLike) -> float:
        try:
            self.evals.value += 1
            y = self.fit(x)
            y0 = y if np.isscalar(y) else sum(y)
            if y0 < self.best_y.value:
                self.best_y.value = y0
                logger.info(str(dtime(self.t0)) + ' '  + 
                              str(self.evals.value) + ' ' + 
                              str(round(self.evals.value/(1E-9 + dtime(self.t0)),0)) + ' ' + 
                              str(self.best_y.value) + ' ' + 
                              str(list(x)))
            return y
        except Exception as ex:
            print(str(ex))  
            return sys.float_info.max  
    
class Optimizer(object):
    """Provides different optimization methods for use with parallel retry."""
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000, 
                 name: Optional[str] = ''):
        self.max_evaluations = max_evaluations  
        self.name = name  

    def max_eval_num(self, store=None):
        return self.max_evaluations if store is None else \
                store.eval_num(self.max_evaluations)
                
    def get_count_runs(self, store=None):
        return 0 if store is None else \
                store.get_count_runs()
                        
class Sequence(Optimizer):
    """Sequence of optimizers."""
    
    def __init__(self, optimizers: ArrayLike):
        Optimizer.__init__(self)
        self.optimizers = optimizers 
        self.max_evaluations = 0 
        for optimizer in self.optimizers:
            self.name += optimizer.name + ' -> '
            self.max_evaluations += optimizer.max_evaluations
        self.name = self.name[:-4]

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Bounds, 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[Union[float, ArrayLike, Callable]] = None, 
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store=None) -> Tuple[np.ndarray, float, int]:
        evals = 0
        y = np.inf
        for optimizer in self.optimizers:
            ret = optimizer.minimize(fun, bounds, guess, sdevs, rg, store)
            if ret[1] < y:
                y = ret[1]
                x = ret[0]
            guess = x
            evals += ret[2]
        return x, y, evals
                  
class Choice(Optimizer):
    """Random choice of optimizers."""
    
    def __init__(self, optimizers: ArrayLike):
        Optimizer.__init__(self)
        self.optimizers = optimizers 
        self.max_evaluations = optimizers[0].max_evaluations 
        for optimizer in self.optimizers:
            self.name += optimizer.name + ' | '
        self.name = self.name[:-3]
                  
    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Bounds, 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[Union[float, ArrayLike, Callable]] = None, 
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store=None) -> Tuple[np.ndarray, float, int]:
        
        choice = rg.integers(0, len(self.optimizers))
        opt = self.optimizers[choice]
        return opt.minimize(fun, bounds, guess, sdevs, rg, store)

def de_cma(max_evaluations: Optional[int] = 50000, 
           popsize: Optional[int] = 31, 
           stop_fitness: Optional[float] = -np.inf, 
           de_max_evals: Optional[int] = None, 
           cma_max_evals: Optional[int] = None, 
           ints: Optional[ArrayLike] = None, 
           workers: Optional[int]  = None) -> Sequence:
    """Sequence differential evolution -> CMA-ES."""

    de_evals = np.random.uniform(0.1, 0.5)
    if de_max_evals is None:
        de_max_evals = int(de_evals*max_evaluations)
    if cma_max_evals is None:
        cma_max_evals = int((1.0-de_evals)*max_evaluations)
    opt1 = De_cpp(popsize=popsize, max_evaluations = de_max_evals, 
                  stop_fitness = stop_fitness, ints=ints, workers = workers)
    opt2 = Cma_cpp(popsize=popsize, max_evaluations = cma_max_evals, 
                   stop_fitness = stop_fitness, workers = workers)
    return Sequence([opt1, opt2])

def de_cma_py(max_evaluations: Optional[int] = 50000, 
           popsize: Optional[int] = 31, 
           stop_fitness: Optional[float] = -np.inf, 
           de_max_evals: Optional[int] = None, 
           cma_max_evals: Optional[int] = None, 
           ints: Optional[ArrayLike] = None, 
           workers: Optional[int]  = None) -> Sequence:
    """Sequence differential evolution -> CMA-ES in python."""

    de_evals = np.random.uniform(0.1, 0.5)
    if de_max_evals is None:
        de_max_evals = int(de_evals*max_evaluations)
    if cma_max_evals is None:
        cma_max_evals = int((1.0-de_evals)*max_evaluations)
    opt1 = De_python(popsize=popsize, max_evaluations = de_max_evals, 
                     stop_fitness = stop_fitness, ints=ints, workers = workers)
    opt2 = Cma_python(popsize=popsize, max_evaluations = cma_max_evals, 
                   stop_fitness = stop_fitness, workers = workers)
    return Sequence([opt1, opt2])

def da_cma(max_evaluations: Optional[int] = 50000, 
           popsize: Optional[int] = 31, 
           da_max_evals: Optional[int] = None, 
           cma_max_evals: Optional[int] = None, 
           stop_fitness: Optional[float] = -np.inf) -> Sequence:
    """Sequence dual annealing -> CMA-ES."""

    da_evals = np.random.uniform(0.1, 0.5)
    if da_max_evals is None:
        da_max_evals = int(da_evals*max_evaluations)
    if cma_max_evals is None:
        cma_max_evals = int((1.0-da_evals)*max_evaluations)
    opt1 = Da_cpp(max_evaluations = da_max_evals, stop_fitness = stop_fitness)
    opt2 = Cma_cpp(popsize=popsize, max_evaluations = cma_max_evals, 
                   stop_fitness = stop_fitness)
    return Sequence([opt1, opt2])

def de_crfmnes(max_evaluations: Optional[int] = 50000, 
               popsize: Optional[int] = 32, 
               stop_fitness: Optional[float] = -np.inf, 
               de_max_evals: Optional[int] = None, 
               crfm_max_evals: Optional[int] = None, 
               ints: Optional[ArrayLike] = None, 
               workers: Optional[int]  = None) -> Sequence:
    """Sequence differential evolution -> CRFMNES."""

    de_evals = np.random.uniform(0.1, 0.5)
    if de_max_evals is None:
        de_max_evals = int(de_evals*max_evaluations)
    if crfm_max_evals is None:
        crfm_max_evals = int((1.0-de_evals)*max_evaluations)
    opt1 = De_cpp(popsize=popsize, max_evaluations = de_max_evals, 
                  stop_fitness = stop_fitness, ints=ints, workers = workers)
    opt2 = Crfmnes_cpp(popsize=popsize, max_evaluations = crfm_max_evals, 
                   stop_fitness = stop_fitness, workers = workers)
    return Sequence([opt1, opt2])

def crfmnes_bite(max_evaluations: Optional[int] = 50000, 
                popsize: Optional[int] = 32, 
                stop_fitness: Optional[float] = -np.inf, 
                crfm_max_evals: Optional[int] = None, 
                bite_max_evals: Optional[int] = None, 
                M: Optional[int] = 1) -> Sequence:
    """Sequence CRFMNES -> Bite."""

    crfmnes_evals = np.random.uniform(0.1, 0.5)
    if crfm_max_evals is None:
        crfm_max_evals = int(crfmnes_evals*max_evaluations)
    if bite_max_evals is None:
        bite_max_evals = int((1.0-crfmnes_evals)*max_evaluations)
    opt1 = Crfmnes_cpp(popsize=popsize, max_evaluations = crfm_max_evals, 
                  stop_fitness = stop_fitness)
    opt2 = Bite_cpp(popsize=popsize, max_evaluations = bite_max_evals, 
                   stop_fitness = stop_fitness, M=M)
    return Sequence([opt1, opt2])

def cma_bite(max_evaluations: Optional[int] = 50000, 
            popsize: Optional[int] = 32, 
            stop_fitness: Optional[float] = -np.inf, 
            cma_max_evals: Optional[int] = None, 
            bite_max_evals: Optional[int] = None, 
            M: Optional[int] = 1) -> Sequence:
    """Sequence CMA-ES -> Bite."""

    cma_evals = np.random.uniform(0.1, 0.5)
    if cma_max_evals is None:
        cma_max_evals = int(cma_evals*max_evaluations)
    if bite_max_evals is None:
        bite_max_evals = int((1.0-cma_evals)*max_evaluations)
    opt1 = Cma_cpp(popsize=popsize, max_evaluations = cma_max_evals, 
                  stop_fitness = stop_fitness, stop_hist = 0)
    opt2 = Bite_cpp(popsize=popsize, max_evaluations = bite_max_evals, 
                   stop_fitness = stop_fitness, M=M)
    return Sequence([opt1, opt2])

class Crfmnes(Optimizer):
    """CRFMNES Python implementation."""
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = 32, 
                 guess: Optional[ArrayLike] = None, 
                 stop_fitness: Optional[float] = -np.inf,
                 sdevs: Optional[float] = None, 
                 workers: Optional[int] = None):        

        Optimizer.__init__(self, max_evaluations, 'crfmnes')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.guess = guess
        self.sdevs = sdevs
        self.workers = workers

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = 0.3, 
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:

        ret = crfmnes.minimize(fun, bounds, 
                self.guess if not self.guess is None else guess,
                input_sigma = self.sdevs if not self.sdevs is None else sdevs,
                max_evaluations = self.max_eval_num(store), 
                popsize=self.popsize, 
                stop_fitness = self.stop_fitness,
                rg=rg, runid=self.get_count_runs(store),
                workers = self.workers)     
        return ret.x, ret.fun, ret.nfev

class Crfmnes_cpp(Optimizer):
    """CRFMNES C++ implementation."""
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = 32, 
                 guess: Optional[ArrayLike] = None, 
                 stop_fitness: Optional[float] = -np.inf,
                 sdevs: Optional[float] = None, 
                 workers: Optional[int] = None):        
       
        Optimizer.__init__(self, max_evaluations, 'crfmnes cpp')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.guess = guess
        self.sdevs = sdevs
        self.workers = workers

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = 0.3, 
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:

        ret = crfmnescpp.minimize(fun, bounds, 
                self.guess if not self.guess is None else guess,
                input_sigma = self.sdevs if not self.sdevs is None else sdevs,
                max_evaluations = self.max_eval_num(store), 
                popsize=self.popsize, 
                stop_fitness = self.stop_fitness,
                rg=rg, runid=self.get_count_runs(store), 
                workers = self.workers)     
        return ret.x, ret.fun, ret.nfev

class Pgpe_cpp(Optimizer):
    """PGPE C++ implementation."""
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 500000,
                 popsize: Optional[int] = 640, 
                 guess: Optional[ArrayLike] = None, 
                 stop_fitness: Optional[float] = -np.inf,
                 sdevs: Optional[float] = None, 
                 workers: Optional[int] = None):
               
        Optimizer.__init__(self, max_evaluations, 'pgpe cpp')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.guess = guess
        self.sdevs = sdevs
        self.workers = workers

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = 0.1, 
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:

        ret = pgpecpp.minimize(fun, bounds, 
                self.guess if not self.guess is None else guess,
                input_sigma = self.sdevs if not self.sdevs is None else sdevs,
                max_evaluations = self.max_eval_num(store), 
                popsize=self.popsize, 
                stop_fitness = self.stop_fitness,
                rg=rg, runid=self.get_count_runs(store), 
                workers = self.workers)     
        return ret.x, ret.fun, ret.nfev

class Cma_python(Optimizer):
    """CMA_ES Python implementation."""
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = 31, 
                 guess: Optional[ArrayLike] = None, 
                 stop_fitness: Optional[float] = -np.inf,
                 sdevs: Optional[float] = None, 
                 workers: Optional[int] = None,        
                 update_gap: Optional[int] = None, 
                 normalize: Optional[bool] = True):  
           
        Optimizer.__init__(self, max_evaluations, 'cma py')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.update_gap = update_gap
        self.guess = guess
        self.sdevs = sdevs
        self.normalize = normalize
        self.workers = workers

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[Union[float, ArrayLike, Callable]] = 0.1, 
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:

        ret = cmaes.minimize(fun, bounds, 
                self.guess if not self.guess is None else guess,
                input_sigma= self.sdevs if not self.sdevs is None else sdevs,
                max_evaluations = self.max_eval_num(store), 
                popsize=self.popsize, 
                stop_fitness = self.stop_fitness,
                rg=rg, runid=self.get_count_runs(store),
                normalize = self.normalize,
                update_gap = self.update_gap,
                workers = self.workers)     
        return ret.x, ret.fun, ret.nfev

class Cma_cpp(Optimizer):
    """CMA_ES C++ implementation."""
   
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = 31, 
                 guess: Optional[ArrayLike] = None, 
                 stop_fitness: Optional[float] = -np.inf,
                 sdevs: Optional[float] = None, 
                 workers: Optional[int] = None,        
                 update_gap: Optional[int] = None, 
                 normalize: Optional[bool] = True,                 
                 delayed_update: Optional[bool] = True,   
                 stop_hist: Optional[int] = None): 
          
        Optimizer.__init__(self, max_evaluations, 'cma cpp')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.stop_hist = stop_hist
        self.guess = guess
        self.sdevs = sdevs
        self.update_gap = update_gap
        self.delayed_update = delayed_update
        self.normalize = normalize
        self.workers = workers

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[Union[float, ArrayLike, Callable]] = 0.1, 
                 rg=Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        
        ret = cmaescpp.minimize(fun, bounds,
                self.guess if not self.guess is None else guess,
                input_sigma = self.sdevs if not self.sdevs is None else sdevs,
                max_evaluations =self.max_eval_num(store),
                popsize = self.popsize,
                stop_fitness = self.stop_fitness,
                stop_hist = self.stop_hist,
                rg = rg, runid = self.get_count_runs(store),
		        update_gap = self.update_gap,
                normalize = self.normalize,
                delayed_update = self.delayed_update,
                workers = self.workers)   
        return ret.x, ret.fun, ret.nfev

class Cma_orig(Optimizer):
    """CMA_ES original implementation."""
   
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = 31, 
                 guess: Optional[ArrayLike] = None, 
                 stop_fitness: Optional[float] = -np.inf,
                 sdevs: Optional[float] = None): 
     
        Optimizer.__init__(self, max_evaluations, 'cma orig')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.guess = guess
        self.sdevs = sdevs

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[Union[float, ArrayLike]] = 0.3, 
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:

        lower = bounds.lb
        upper = bounds.ub
        guess = self.guess if not self.guess is None else guess
        if guess is None:
            guess = rg.uniform(lower, upper)
        max_evaluations = self.max_eval_num(store)   
        input_sigma= self.sdevs if not self.sdevs is None else sdevs
        try:
            import cma
        except ImportError as e:
            raise ImportError("Please install CMA (pip install cma)") 
        try: 
            es = cma.CMAEvolutionStrategy(guess, 0.1,  {'bounds': [lower, upper], 
                                                             'typical_x': guess,
                                                             'scaling_of_variables': scale(lower, upper),
                                                             'popsize': self.popsize,
                                                             'CMA_stds': input_sigma,
                                                             'verbose': -1,
                                                             'verb_disp': -1})
            evals = 0
            for i in range(max_evaluations):
                X, Y = es.ask_and_eval(fun)
                es.tell(X, Y)
                evals += self.popsize
                if es.stop():
                    break 
                if evals > max_evaluations:
                    break    
            return es.result.xbest, es.result.fbest, evals
        except Exception as ex:
            print(ex)

class Cma_lw(Optimizer):
    """CMA lightweight Python implementation. See https://github.com/CyberAgentAILab/cmaes """
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = 31, 
                 guess: Optional[ArrayLike] = None, 
                 stop_fitness: Optional[float] = -np.inf,
                 sdevs: Optional[Union[float, ArrayLike]] = None, 
                 workers: Optional[int] = None):
      
        Optimizer.__init__(self, max_evaluations, 'cma_lw')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.guess = guess
        self.sdevs = sdevs
        self.workers = workers

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[Union[float, ArrayLike]] = 0.3, 
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        
        try:
            import cmaes
        except ImportError as e:
            raise ImportError("Please install cmaes (pip install cmaes)") 

        if guess is None:
            guess = self.guess
        if guess is None:    
            guess = rg.uniform(bounds.lb, bounds.ub)
        bds = np.array([t for t in zip(bounds.lb, bounds.ub)])
        seed = int(rg.uniform(0, 2**32 - 1))
        optimizer = cmaes.CMA(mean=guess, sigma=np.mean(sdevs), bounds=bds, seed=seed, population_size=self.popsize)
        best_y = np.inf
        evals = 0
        fun = serial(fun) if (self.workers is None or self.workers <= 1) else parallel(fun, self.workers)  
        while evals < self.max_evaluations and not optimizer.should_stop():
            xs = [optimizer.ask() for _ in range(optimizer.population_size)]
            ys = fun(xs)
            solutions = []
            for i in range(optimizer.population_size):
                x = xs[i]
                y = ys[i]
                solutions.append((x, y))
                if y < best_y:
                    best_y = y
                    best_x = x
            optimizer.tell(solutions)
            evals += optimizer.population_size           
        if isinstance(fun, parallel):
            fun.stop()
        return best_x, best_y, evals

class Cma_awm(Optimizer):
    """CMA awm Python implementation. See https://github.com/CyberAgentAILab/cmaes """
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = 31, 
                 guess: Optional[ArrayLike] = None, 
                 stop_fitness: Optional[float] = -np.inf,
                 sdevs: Optional[Union[float, ArrayLike]] = None, 
                 continuous_space = None, 
                 discrete_space = None, 
                 workers: Optional[int] = None):
               
        Optimizer.__init__(self, max_evaluations, 'cma_awm')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.guess = guess
        self.sdevs = sdevs
        self.workers = workers
        self.continuous_space = continuous_space
        self.discrete_space = discrete_space

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[Union[float, ArrayLike]] = 0.3, 
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        try:
            import cmaes
        except ImportError as e:
            raise ImportError("Please install cmaes (pip install cmaes)") 
              
        if guess is None:
            guess = self.guess
        if guess is None:    
            guess = rg.uniform(bounds.lb, bounds.ub)
        seed = int(rg.uniform(0, 2**32 - 1))
        optimizer = cmaes.CMAwM(mean=guess, sigma=np.mean(sdevs),       
                         continuous_space=self.continuous_space,  
                         discrete_space=self.discrete_space, 
                         seed=seed, population_size=self.popsize)
        best_y = 1E99
        evals = 0
        fun = serial(fun) if (self.workers is None or self.workers <= 1) else parallel(fun, self.workers)  
        while evals < self.max_evaluations and not optimizer.should_stop():
            asks = [optimizer.ask() for _ in range(optimizer.population_size)]
            ys = fun([x[0] for x in asks])
            solutions = []
            for i in range(optimizer.population_size):
                x = asks[i][1]
                y = ys[i]
                solutions.append((x, y))
                if y < best_y:
                    best_y = y
                    best_x = x
            optimizer.tell(solutions)
            evals += optimizer.population_size           
        if isinstance(fun, parallel):
            fun.stop()
        return best_x, best_y, evals

class Cma_sep(Optimizer):
    """CMA sep Python implementation. See https://github.com/CyberAgentAILab/cmaes """
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = 31, 
                 guess: Optional[ArrayLike] = None, 
                 stop_fitness: Optional[float] = -np.inf,
                 sdevs: Optional[Union[float, ArrayLike]] = None, 
                 workers: Optional[int] = None):
      
        Optimizer.__init__(self, max_evaluations, 'cma_sep')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.guess = guess
        self.sdevs = sdevs
        self.workers = workers

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[Union[float, ArrayLike]] = 0.3, 
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        try:
            import cmaes
        except ImportError as e:
            raise ImportError("Please install cmaes (pip install cmaes)") 

        if guess is None:
            guess = self.guess
        if guess is None:    
            guess = rg.uniform(bounds.lb, bounds.ub)
        bds = np.array([t for t in zip(bounds.lb, bounds.ub)])
        seed = int(rg.uniform(0, 2**32 - 1))
        optimizer = cmaes.SepCMA(mean=guess, sigma=np.mean(sdevs), bounds=bds, seed=seed, population_size=self.popsize)
        best_y = np.inf
        evals = 0
        fun = serial(fun) if (self.workers is None or self.workers <= 1) else parallel(fun, self.workers)  
        while evals < self.max_evaluations and not optimizer.should_stop():
            xs = [optimizer.ask() for _ in range(optimizer.population_size)]
            ys = fun(xs)
            solutions = []
            for i in range(optimizer.population_size):
                x = xs[i]
                y = ys[i]
                solutions.append((x, y))
                if y < best_y:
                    best_y = y
                    best_x = x
            optimizer.tell(solutions)
            evals += optimizer.population_size          
        if isinstance(fun, parallel):
            fun.stop()
        return best_x, best_y, evals
      
class De_cpp(Optimizer):
    """Differential Evolution C++ implementation."""
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = None, 
                 stop_fitness: Optional[float] = -np.inf,
                 keep: Optional[int] = 200, 
                 f: Optional[float] = 0.5, 
                 cr: Optional[float] = 0.9, 
                 ints: Optional[ArrayLike] = None, 
                 workers: Optional[int] = None):
      
        Optimizer.__init__(self, max_evaluations, 'de cpp')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.keep = keep
        self.f = f
        self.cr = cr
        self.ints = ints
        self.workers = workers

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None,  # ignored
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:

        ret = decpp.minimize(fun, None, bounds, 
                popsize=self.popsize, 
                max_evaluations = self.max_eval_num(store), 
                stop_fitness = self.stop_fitness,
                keep = self.keep, f = self.f, cr = self.cr, ints=self.ints,
                rg=rg, runid = self.get_count_runs(store), 
                workers = self.workers)
        return ret.x, ret.fun, ret.nfev

class De_python(Optimizer):
    """Differential Evolution Python implementation."""
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = None,
                 stop_fitness: Optional[float] = -np.inf,
                 keep: Optional[int] = 200, 
                 f: Optional[float] = 0.5, 
                 cr: Optional[float] = 0.9, 
                 ints: Optional[ArrayLike] = None, 
                 workers: Optional[int] = None):
             
        Optimizer.__init__(self, max_evaluations, 'de py')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.keep = keep
        self.f = f
        self.cr = cr
        self.ints = ints
        self.workers = workers

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        
        ret = de.minimize(fun, None, 
                bounds, self.popsize, self.max_eval_num(store),
                stop_fitness = self.stop_fitness,
                keep = self.keep, f = self.f, cr = self.cr, ints=self.ints,
                rg=rg, workers = self.workers)
        return ret.x, ret.fun, ret.nfev

class Cma_ask_tell(Optimizer):
    """CMA ask tell implementation."""
    
    def __init__(self, max_evaluations=50000,
                 popsize = 31, guess=None, stop_fitness = -np.inf, sdevs = None):        
        Optimizer.__init__(self, max_evaluations, 'cma at')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.guess = guess
        self.sdevs = sdevs

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        
        es = cmaes.Cmaes(bounds,
                popsize = self.popsize, 
                input_sigma = self.sdevs if not self.sdevs is None else sdevs, 
                rg = rg)       
        iters = self.max_eval_num(store) // self.popsize
        evals = 0
        for j in range(iters):
            xs = es.ask()
            ys = [fun(x) for x in xs]
            evals += len(xs)
            stop = es.tell(ys)
            if stop != 0:
                break 
        return es.best_x, es.best_value, evals

class De_ask_tell(Optimizer):
    """Differential Evolution ask tell implementation."""
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = None,
                 stop_fitness: Optional[float] = -np.inf,
                 keep: Optional[int] = 200, 
                 f: Optional[float] = 0.5, 
                 cr: Optional[float] = 0.9):        
        Optimizer.__init__(self, max_evaluations, 'de at')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.keep = keep
        self.f = f
        self.cr = cr

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        
        dim = len(bounds.lb)
        popsize = 31 if self.popsize is None else self.popsize
        es = de.DE(dim, bounds, popsize = popsize, rg = rg, keep = self.keep, F = self.f, Cr = self.cr)  
        es.fun = fun  #remove
        max_evals = self.max_eval_num(store)
        while es.evals < max_evals:
            xs = es.ask()
            ys = [fun(x) for x in xs]
            stop = es.tell(ys, xs)
            if stop != 0:
                break 
        return es.best_x, es.best_value, es.evals

class random_search(Optimizer):
    """Random search."""
   
    def __init__(self, max_evaluations=50000):        
        Optimizer.__init__(self, max_evaluations, 'random')
 
    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        
        dim, x_min, y_min = len(bounds.lb), None, None
        max_chunk_size = 1 + 4e4 / dim
        evals = self.max_eval_num(store)
        budget = evals
        while budget > 0:
            chunk = int(max([1, min([budget, max_chunk_size])]))
            X = rg.uniform(bounds.lb, bounds.ub, size = [chunk, dim])
            F = [fun(x) for x in X]
            index = np.argmin(F) if len(F) else None
            if index is not None and (y_min is None or F[index] < y_min):
                x_min, y_min = X[index], F[index]
            budget -= chunk
        return x_min, y_min, evals

class LDe_cpp(Optimizer):
    """Local Differential Evolution C++ implementation."""
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = None,
                 stop_fitness: Optional[float] = -np.inf,
                 keep: Optional[int] = 200, 
                 f: Optional[float] = 0.5, 
                 cr: Optional[float] = 0.9, 
                 guess: Optional[ArrayLike] = None, 
                 sdev: Optional[Union[float, ArrayLike, Callable]] = None, 
                 ints: Optional[ArrayLike] = None):
               
        Optimizer.__init__(self, max_evaluations, 'lde cpp')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.keep = keep
        self.f = f
        self.cr = cr
        self.guess = guess
        self.sdevs = sdev
        self.ints = ints
        
    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        
        ret = ldecpp.minimize(fun, bounds, 
                self.guess if not self.guess is None else guess, 
                self.sdevs if not self.sdevs is None else sdevs,
                popsize=self.popsize, 
                max_evaluations = self.max_eval_num(store), 
                stop_fitness = self.stop_fitness,
                keep = self.keep, f = self.f, cr = self.cr, ints = self.ints,
                rg=rg, runid = self.get_count_runs(store))
        return ret.x, ret.fun, ret.nfev

class GCLDE_cpp(Optimizer):
    """GCL-Differential Evolution C++ implementation."""
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = None,
                 stop_fitness: Optional[float] = -np.inf,
                 pbest: Optional[float] = 0.7, 
                 f: Optional[float] = 0.5, 
                 cr: Optional[float] = 0.9, 
                 workers: Optional[int] = None):
                
        Optimizer.__init__(self, max_evaluations, 'gclde cpp')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.pbest = pbest
        self.f = f
        self.cr = cr
        self.workers = workers

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg: Optional[Generator] = Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        
        ret = gcldecpp.minimize(fun, None, bounds, 
                popsize=self.popsize, 
                max_evaluations = self.max_eval_num(store), 
                stop_fitness = self.stop_fitness,
                pbest = self.pbest, f0 = self.f, cr0 = self.cr,
                rg=rg, runid = self.get_count_runs(store),
                workers = self.workers)
        return ret.x, ret.fun, ret.nfev

class LCLDE_cpp(Optimizer):
    """LCL-Differential Evolution C++ implementation."""
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = None,
                 stop_fitness: Optional[float] = -np.inf,
                 pbest: Optional[float] = 0.7, 
                 f: Optional[float] = 0.5, 
                 cr: Optional[float] = 0.9, 
                 guess: Optional[ArrayLike] = None, 
                 sdev: Optional[Union[float, ArrayLike, Callable]] = None, 
                 workers: Optional[int] = None):                 
       
        Optimizer.__init__(self, max_evaluations, 'lclde cpp')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.pbest = pbest
        self.f = f
        self.cr = cr
        self.workers = workers
        self.guess = guess
        self.sdevs = sdev

    def minimize(self, 
                fun: Callable[[ArrayLike], float], 
                bounds: Optional[Bounds], 
                guess: Optional[ArrayLike] = None, 
                sdevs: Optional[float] = 0.3, # ignored
                rg=Generator(MT19937()), 
                store = None) -> Tuple[np.ndarray, float, int]:                 
        
        ret = lcldecpp.minimize(fun, bounds, 
                self.guess if not self.guess is None else guess, 
                self.sdevs if not self.sdevs is None else sdevs,
                popsize=self.popsize, 
                max_evaluations = self.max_eval_num(store), 
                stop_fitness = self.stop_fitness,
                pbest = self.pbest, f0 = self.f, cr0 = self.cr,
                rg=rg, runid = self.get_count_runs(store),
                workers = self.workers)

        return ret.x, ret.fun, ret.nfev
    
class Da_cpp(Optimizer):
    """Dual Annealing C++ implementation."""
    
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 stop_fitness: Optional[float] = -np.inf,
                 use_local_search: Optional[bool] = True,
                 guess: Optional[ArrayLike] = None):                 
                
        Optimizer.__init__(self, max_evaluations, 'da cpp',)
        self.stop_fitness = stop_fitness
        self.use_local_search = use_local_search
        self.guess = guess
 
    def minimize(self, 
                fun: Callable[[ArrayLike], float], 
                bounds: Optional[Bounds], 
                guess: Optional[ArrayLike] = None, 
                sdevs: Optional[float] = None, # ignored
                rg=Generator(MT19937()), 
                store = None) -> Tuple[np.ndarray, float, int]:

        ret = dacpp.minimize(fun, bounds, 
                             self.guess if guess is None else guess,
                            max_evaluations = self.max_eval_num(store), 
                            use_local_search = self.use_local_search,
                            rg=rg, runid = self.get_count_runs(store))
        return ret.x, ret.fun, ret.nfev

class Csma_cpp(Optimizer):
    """SCMA C++ implementation."""
   
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = None,
                 guess: Optional[ArrayLike] = None, 
                 stop_fitness: Optional[float] = -np.inf,
                 sdevs: Optional[float] = None):
                
        Optimizer.__init__(self, max_evaluations, 'scma cpp')
        self.popsize = popsize
        self.stop_fitness = stop_fitness
        self.guess = guess
        self.sdevs = sdevs

    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = 0.16, 
                 rg=Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        
        ret = csmacpp.minimize(fun, bounds, 
                self.guess if guess is None else guess,
                self.sdevs if not self.sdevs is None else sdevs,
                max_evaluations = self.max_eval_num(store), 
                stop_fitness = self.stop_fitness,
                rg=rg, runid = self.get_count_runs(store))     
        return ret.x, ret.fun, ret.nfev

class Bite_cpp(Optimizer):
    """Bite C++ implementation."""
   
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 guess: Optional[ArrayLike] = None, 
                 stop_fitness: Optional[float] = -np.inf,                
                 M: Optional[int] = None,
                 popsize: Optional[int] = None,
                 stall_criterion: Optional[int] = None):
                
        Optimizer.__init__(self, max_evaluations, 'bite cpp')
        self.guess = guess
        self.stop_fitness = stop_fitness
        self.M = 1 if M is None else M 
        self.popsize = 0 if popsize is None else popsize 
        self.stall_criterion = 0 if stall_criterion is None else stall_criterion 

    def minimize(self,
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg=Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:

        ret = bitecpp.minimize(fun, bounds, 
                self.guess if guess is None else guess,
                max_evaluations = self.max_eval_num(store), 
                stop_fitness = self.stop_fitness, M = self.M, popsize = self.popsize, 
                stall_criterion = self.stall_criterion,
                rg=rg, runid = self.get_count_runs(store))     
        return ret.x, ret.fun, ret.nfev
        
class Dual_annealing(Optimizer):
    """scipy dual_annealing."""
 
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 use_local_search: Optional[bool] = True):
                
        Optimizer.__init__(self, max_evaluations, 'scipy da')
        self.no_local_search = not use_local_search
 
    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg=Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:                 
        ret = dual_annealing(fun, bounds=list(zip(bounds.lb, bounds.ub)),
            maxfun = self.max_eval_num(store), 
            no_local_search = self.no_local_search,
            x0=guess,
            seed = int(rg.uniform(0, 2**32 - 1)))
        return ret.x, ret.fun, ret.nfev

class Differential_evolution(Optimizer):
    """scipy differential_evolution."""
 
    def __init__(self, 
                 max_evaluations: Optional[int] = 50000,
                 popsize: Optional[int] = 31):
                
        Optimizer.__init__(self, max_evaluations, 'scipy de')
        self.popsize = popsize
 
    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg=Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        
        popsize = self.popsize 
        maxiter = int(self.max_eval_num(store) / (popsize * len(bounds.lb)) - 1)
        ret = differential_evolution(fun, bounds=bounds, maxiter=maxiter,
                                      seed = int(rg.uniform(0, 2**32 - 1)))
        return ret.x, ret.fun, ret.nfev

class CheckBounds(object):
    
    def __init__(self, bounds: Bounds):
        self.bounds = bounds
        
    def __call__(self, **kwargs):
        x = kwargs["x_new"]
        inb =  np.less_equal(x, self.bounds.ub).all() and \
            np.greater_equal(x, self.bounds.lb).all()
        return inb

class Basin_hopping(Optimizer):
    """scipy basin hopping."""
 
    def __init__(self, max_evaluations=50000, store=None):        
        Optimizer.__init__(self, max_evaluations, 'scipy basin hopping')
         
    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg=Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        localevals = 200
        maxiter = int(self.max_eval_num(store) / localevals)         
        if guess is None:
            guess = rg.uniform(bounds.lb, bounds.ub)
        
        ret = basinhopping(fun, guess, niter=maxiter, 
                           minimizer_kwargs={"method": 'SLSQP', 
                                             "bounds":bounds},
                           accept_test=CheckBounds(bounds),
                           seed = int(rg.uniform(0, 2**32 - 1)))
        return ret.x, ret.fun, ret.nfev

class Minimize(Optimizer):
    """scipy minimize."""
 
    def __init__(self, max_evaluations=50000, store=None):        
        Optimizer.__init__(self, max_evaluations, 'scipy minimize')
 
    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg=Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        
        if guess is None:
            guess = rg.uniform(bounds.lb, bounds.ub)
        ret = minimize(fun, x0=guess, bounds=bounds)
        return ret.x, ret.fun, ret.nfev
 
class Shgo(Optimizer):
    """scipy shgo."""

    def __init__(self, max_evaluations=50000, store=None):        
        Optimizer.__init__(self, max_evaluations, 'scipy shgo')
 
    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg=Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:
        
        ret = shgo(fun, bounds=list(zip(bounds.lb, bounds.ub)), 
                   options={'maxfev': self.max_eval_num(store)})
        return ret.x, ret.fun, ret.nfev

class single_objective:
    """Utility class to create a fcmaes problem from a pagmo problem."""
      
    def __init__(self, pagmo_prob):
        self.pagmo_prob = pagmo_prob
        self.name = pagmo_prob.get_name() 
        self.fun = self.fitness
        lb, ub = pagmo_prob.get_bounds()
        self.bounds = Bounds(lb, ub)
         
    def fitness(self,X):
        try:
            return self.pagmo_prob.fitness(X)[0]
        except Exception as ex:
            return sys.float_info.max

class NLopt(Optimizer):
    """NLopt_algo wrapper."""

    def __init__(self, algo, max_evaluations=50000, store=None):        
        Optimizer.__init__(self, max_evaluations, 'NLopt ' + algo.get_algorithm_name())
        self.algo = algo
 
    def minimize(self, 
                 fun: Callable[[ArrayLike], float], 
                 bounds: Optional[Bounds], 
                 guess: Optional[ArrayLike] = None, 
                 sdevs: Optional[float] = None, # ignored
                 rg=Generator(MT19937()), 
                 store = None) -> Tuple[np.ndarray, float, int]:

        self.fun = fun
        opt = self.algo
        opt.set_min_objective(self.nlfunc)
        opt.set_lower_bounds(bounds.lb)
        opt.set_upper_bounds(bounds.ub)
        opt.set_maxeval(self.max_eval_num(store))
        opt.set_initial_step(sdevs)
        if guess is None:
            guess = rg.uniform(bounds.lb, bounds.ub)
        x = opt.optimize(guess)
        y = opt.last_optimum_value()
        return x, y, opt.get_numevals()
    
    def nlfunc(self, x, _):
        try:
            return self.fun(x)
        except Exception as ex:
            return sys.float_info.max
    



# Copyright (c) Dietmar Wolz.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory.
from __future__ import annotations

""" Parallel objective function evaluator.
    Uses pipes to avoid re-spawning new processes for each eval_parallel call. 
    the objective function is distributed once to all processes and
    reused for all eval_parallel calls. Evaluator(fun) needs to be stopped after the
    whole optimization is finished to avoid a resource leak.
"""

from multiprocessing import Process, Pipe
import multiprocessing as mp
import ctypes as ct
import numpy as np
import sys, math, os  
from loguru import logger
from typing import Optional, Callable, Tuple
from numpy.typing import ArrayLike

pipe_limit = 64 # higher values can cause issues

def is_log_level_active(level):
    try: # nasty but currently there is no other way
        for handler in logger._core.handlers.values():
            if handler._levelno <= logger.level(level).no:
                return True
    except Exception as ex:   
        pass
    return False

def is_debug_active():
    return is_log_level_active("DEBUG")

def is_trace_active():
    return is_log_level_active("TRACE")

def eval_parallel(xs: ArrayLike, 
                  evaluator: Evaluator):
    popsize = len(xs)
    ys = np.empty(popsize)
    i0 = 0
    i1 = min(popsize, pipe_limit)
    while True:
        _eval_parallel_segment(xs, ys, i0, i1, evaluator)
        if i1 >= popsize:
            break;
        i0 += pipe_limit
        i1 = min(popsize, i1 + pipe_limit)
    return ys
        
def eval_parallel_mo(xs: ArrayLike, 
                     evaluator: Evaluator, 
                     nobj: int):
    popsize = len(xs)
    ys = np.empty((popsize,nobj))
    i0 = 0
    i1 = min(popsize, pipe_limit)
    while True:
        _eval_parallel_segment(xs, ys, i0, i1, evaluator)
        if i1 >= popsize:
            break;
        i0 += pipe_limit
        i1 = min(popsize, i1 + pipe_limit)
    return ys
        
class Evaluator(object):
       
    def __init__(self, 
                 fun: Callable[[ArrayLike], float], # objective function
                ):   
        self.fun = fun 
        self.pipe = Pipe()
        self.read_mutex = mp.Lock() 
        self.write_mutex = mp.Lock() 
            
    def start(self, workers: Optional[int] = mp.cpu_count()):
        self.workers = workers
        self.proc=[Process(target=_evaluate, args=(self.fun, 
                self.pipe, self.read_mutex, self.write_mutex)) for _ in range(workers)]
        [p.start() for p in self.proc]
        
    def stop(self): # shutdown all workers 
        for _ in range(self.workers):
            self.pipe[0].send(None)
        [p.join() for p in self.proc]    
        for p in self.pipe:
            p.close()

def _eval_parallel_segment(xs, ys, i0, i1, evaluator):
    for i in range(i0, i1):
        evaluator.pipe[0].send((i, xs[i]))
    for _ in range(i0, i1):        
        i, y = evaluator.pipe[0].recv()
        ys[i] = y
    return ys

def _evaluate(fun, pipe, read_mutex, write_mutex): # worker
    while True:
        with read_mutex:
            msg = pipe[1].recv() # Read from the input pipe
        if msg is None: 
            break # shutdown worker
        try:
            i, x = msg
            y = fun(x)
        except Exception as ex:
            y =  sys.float_info.max
        with write_mutex:            
            pipe[1].send((i, y)) # Send result

def _check_bounds(bounds, guess, rg):
    if bounds is None and guess is None:
        raise ValueError('either guess or bounds need to be defined')
    if bounds is None:
        return None, None, np.asarray(guess)
    if guess is None:
        guess = rg.uniform(bounds.lb, bounds.ub)
    return np.asarray(bounds.lb), np.asarray(bounds.ub), np.asarray(guess)

def _get_bounds(dim, bounds, guess, rg):
    if bounds is None:
        if guess is None:
            guess = np.asarray(np.zeros(dim))
        return None, None, guess
    if guess is None:
        guess = rg.uniform(bounds.lb, bounds.ub)
    return np.asarray(bounds.lb), np.asarray(bounds.ub), np.asarray(guess)

class _fitness(object):
    """wrapper around the objective function, scales relative to boundaries."""
     
    def __init__(self, fun, lower, upper, normalize = None):
        self.fun = fun
        self.evaluation_counter = 0
        self.lower = lower
        self.normalize = False
        if not (lower is None or normalize is None):
            self.normalize = normalize
        if not lower is None:
            self.upper = upper
            self.scale = 0.5 * (upper - lower)
            self.typx = 0.5 * (upper + lower)

    def values(self, Xs): #enables parallel evaluation
        values = self.fun(Xs)
        self.evaluation_counter += len(Xs)
        return np.array(values)
    
    def closestFeasible(self, X):
        if self.lower is None:
            return X    
        else:
            if self.normalize:
                return np.clip(X, -1.0, 1.0)
            else:
                return np.clip(X, self.lower, self.upper)

    def encode(self, X):
        if self.normalize:
            return (X - self.typx) / self.scale
        else:
            return X
   
    def decode(self, X):
        if self.normalize:
            return (X * self.scale) + self.typx
        else:
            return X
         
def serial(fun):
    """Convert an objective function for serial execution for cmaes.minimize.
    
    Parameters
    ----------
    fun : objective function mapping a list of float arguments to a float value

    Returns
    -------
    out : function
        A function mapping a list of lists of float arguments to a list of float values
        by applying the input function in a loop."""
  
    return lambda xs : [_tryfun(fun, x) for x in xs]
        
def _func_serial(fun, num, pid, xs, ys):
    for i in range(pid, len(xs), num):
        ys[i] = _tryfun(fun, xs[i])

def _tryfun(fun, x):
    try:
        fit = fun(x)
        return fit if math.isfinite(fit) else sys.float_info.max
    except Exception:
        return sys.float_info.max
    
class parallel(object):
    """Convert an objective function for parallel execution for cmaes.minimize.
    
    Parameters
    ----------
    fun : objective function mapping a list of float arguments to a float value.
   
    represents a function mapping a list of lists of float arguments to a list of float values
    by applying the input function using parallel processes. stop needs to be called to avoid
    a resource leak"""
        
    def __init__(self, 
                 fun: Callable[[ArrayLike], float], 
                 workers: Optional[int] = mp.cpu_count()):
        self.evaluator = Evaluator(fun)
        self.evaluator.start(workers)
    
    def __call__(self, xs: ArrayLike) -> np.ndarray:
        return eval_parallel(xs, self.evaluator)

    def stop(self):
        self.evaluator.stop()

class parallel_mo(object):
        
    def __init__(self, 
                 fun: Callable[[ArrayLike], ArrayLike], 
                 nobj: int, 
                 workers: Optional[int] = mp.cpu_count()):
        self.nobj = nobj
        self.evaluator = Evaluator(fun)
        self.evaluator.start(workers)
    
    def __call__(self, xs: ArrayLike) -> np.ndarray:
        return eval_parallel_mo(xs, self.evaluator, self.nobj)

    def stop(self):
        self.evaluator.stop()

class callback(object):
    
    def __init__(self, fun: Callable[[ArrayLike], float]):
        self.fun = fun
    
    def __call__(self, n: int, x: ArrayLike) -> float:
        try:
            fit = self.fun(np.fromiter((x[i] for i in range(n)), dtype=float))
            return fit if math.isfinite(fit) else sys.float_info.max
        except Exception as ex:
            return sys.float_info.max
        
class callback_so(object):
    
    def __init__(self, 
                 fun: Callable[[ArrayLike], float], 
                 dim: int, 
                 is_terminate: Optional[Callable[[ArrayLike, float], bool]] = None):
        self.fun = fun
        self.dim = dim
        self.nobj = 1
        self.is_terminate = is_terminate
    
    def __call__(self, dim, x, y):
        try:
            arrTypeX = ct.c_double*(self.dim)
            xaddr = ct.addressof(x.contents)
            xbuf = np.frombuffer(arrTypeX.from_address(xaddr))
            arrTypeY = ct.c_double*(self.nobj)
            yaddr = ct.addressof(y.contents)   
            ybuf = np.frombuffer(arrTypeY.from_address(yaddr))  
            fit = self.fun(xbuf)
            ybuf[0] = fit if math.isfinite(fit) else sys.float_info.max
            return False if self.is_terminate is None else self.is_terminate(xbuf, ybuf) 
        except Exception as ex:
            print (ex)
            return False

class callback_mo(object):
    
    def __init__(self, 
                 fun: Callable[[ArrayLike], ArrayLike], 
                 dim: int, 
                 nobj: int, 
                 is_terminate: Optional[bool] = None):
        self.fun = fun
        self.dim = dim
        self.nobj = nobj
        self.is_terminate = is_terminate
    
    def __call__(self, dim: int, x, y):
        try:
            arrTypeX = ct.c_double*(dim)
            xaddr = ct.addressof(x.contents)
            xbuf = np.frombuffer(arrTypeX.from_address(xaddr))
            arrTypeY = ct.c_double*(self.nobj)
            yaddr = ct.addressof(y.contents)   
            ybuf = np.frombuffer(arrTypeY.from_address(yaddr))  
            ybuf[:] = self.fun(xbuf)[:]
            return False if self.is_terminate is None else self.is_terminate(xbuf, ybuf) 
        except Exception as ex:
            print (ex)
            return False

class callback_par(object):
    
    def __init__(self, 
                 fun: Callable[[ArrayLike], float], 
                 parfun: Callable[[ArrayLike], ArrayLike]):
        self.fun = fun
        self.parfun = parfun
    
    def __call__(self, popsize, n, xs_, ys_):
        try:
            arrType = ct.c_double*(popsize*n)
            addr = ct.addressof(xs_.contents)
            xall = np.frombuffer(arrType.from_address(addr))
            
            if self.parfun is None:
                for p in range(popsize):
                    ys_[p] = self.fun(xall[p*n : (p+1)*n])
            else:    
                xs = []
                for p in range(popsize):
                    x = xall[p*n : (p+1)*n]
                    xs.append(x)
                ys = self.parfun(xs)
                for p in range(popsize):
                    ys_[p] = ys[p]
        except Exception as ex:
            print (ex)

basepath = os.path.dirname(os.path.abspath(__file__))

try: 
    if sys.platform.startswith('linux'):
        libcmalib = ct.cdll.LoadLibrary(basepath + '/lib/libacmalib.so')  
    elif 'mac' in sys.platform or 'darwin' in sys.platform:
        libcmalib = ct.cdll.LoadLibrary(basepath + '/lib/libacmalib.dylib')  
    else:
        os.environ['PATH'] = (basepath + '/lib') + os.pathsep + os.environ['PATH']
        libcmalib = ct.cdll.LoadLibrary(basepath + '/lib/libacmalib.dll')
except Exception as ex:
    libcmalib = None
    
mo_call_back_type = ct.CFUNCTYPE(ct.c_bool, ct.c_int, ct.POINTER(ct.c_double), ct.POINTER(ct.c_double))
  
call_back_type = ct.CFUNCTYPE(ct.c_double, ct.c_int, ct.POINTER(ct.c_double))  

call_back_par = ct.CFUNCTYPE(None, ct.c_int, ct.c_int, \
                                  ct.POINTER(ct.c_double), ct.POINTER(ct.c_double))  





import math
import numpy as np
import os
from scipy.optimize import OptimizeResult, Bounds
from numpy.random import MT19937, Generator
from fcmaes.evaluator import _get_bounds, _fitness, serial, parallel

from typing import Optional, Callable, Union, Dict
from numpy.typing import ArrayLike

""" Numpy based implementation of Fast Moving Natural Evolution Strategy 
    for High-Dimensional Problems (CR-FM-NES), see https://arxiv.org/abs/2201.11422 .
    Derived from https://github.com/nomuramasahir0/crfmnes .
"""

# evaluation value of the infeasible solution
INFEASIBLE = np.inf

os.environ['MKL_DEBUG_CPU_TYPE'] = '5'

def minimize(fun: Callable[[ArrayLike], float],
             bounds: Optional[Bounds] = None,
             x0: Optional[ArrayLike] = None,
             input_sigma: Optional[float] = 0.3,
             popsize: Optional[int] = 32,
             max_evaluations: Optional[int] = 100000,
             workers: Optional[int] = None,
             stop_fitness: Optional[float] = -np.inf,
             is_terminate: Optional[Callable[[ArrayLike, float], bool]] = None,
             rg: Optional[Generator] = Generator(MT19937()),
             runid: Optional[int] = 0,
             normalize: Optional[bool] = False,
             options: Optional[Dict] = {}
             ) -> OptimizeResult:       
    """Minimization of a scalar function of one or more variables using CMA-ES.
     
    Parameters
    ----------
    fun : callable
        The objective function to be minimized.
            ``fun(x) -> float``
        where ``x`` is an 1-D array with shape (n,)
    bounds : sequence or `Bounds`, optional
        Bounds on variables. There are two ways to specify the bounds:
            1. Instance of the `scipy.Bounds` class.
            2. Sequence of ``(min, max)`` pairs for each element in `x`. None
               is used to specify no bound.
    x0 : ndarray, shape (n,)
        Initial guess. Array of real elements of size (n,),
        where 'n' is the number of independent variables.  
    input_sigma : ndarray, shape (n,) or scalar
        Initial step size.
    popsize = int, optional
        CMA-ES population size.
    max_evaluations : int, optional
        Forced termination after ``max_evaluations`` function evaluations.
    workers : int or None, optional
        If not workers is None, function evaluation is performed in parallel for the whole population. 
        Useful for costly objective functions but is deactivated for parallel retry.     
    stop_fitness : float, optional 
         Limit for fitness value. If reached minimize terminates.
    is_terminate : callable, optional
        Callback to be used if the caller of minimize wants to 
        decide when to terminate. 
    rg = numpy.random.Generator, optional
        Random generator for creating random guesses.
    runid : int, optional
        id used by the is_terminate callback to identify the optimization run.
    normalize : boolean, optional
        if true pheno -> geno transformation maps arguments to interval [-1,1]  
    options : dict, optional
   
    Returns
    -------
    res : scipy.OptimizeResult
        The optimization result is represented as an ``OptimizeResult`` object"""

    cr = CRFMNES(None, bounds, x0, input_sigma, popsize, 
                 max_evaluations, stop_fitness, is_terminate, runid, normalize, options, rg, workers, fun)
    
    cr.optimize()

    return OptimizeResult(x=cr.f.decode(cr.x_best), fun=cr.f_best, nfev=cr.no_of_evals, 
                          nit=cr.g, status=cr.stop, 
                          success=True)

class CRFMNES:
    
    def __init__(self, 
                dim = None, 
                bounds: Optional[Bounds] = None, 
                x0: Optional[ArrayLike] = None,
                input_sigma: Optional[Union[float, ArrayLike, Callable]] = 0.3, 
                popsize: Optional[int] = 32,  
                max_evaluations: Optional[int] = 100000, 
                stop_fitness: Optional[float] = -np.inf, 
                is_terminate: Optional[bool] = None, 
                runid: Optional[int] = 0, 
                normalize: Optional[bool] = False,
                options: Optional[Dict] = {}, 
                rg: Optional[Generator] = Generator(MT19937()), 
                workers: Optional[int] = None, 
                fun: Optional[Callable[[ArrayLike], float]] = lambda x: 0): 
        
        if popsize is None:
            popsize = 32         
        if popsize % 2 == 1: # requires even popsize
            popsize += 1
        if dim is None:
            if not x0 is None: dim = len(x0)
            else: 
                if not bounds is None: dim = len(bounds.lb)
        lower, upper, guess = _get_bounds(dim, bounds, x0, rg) 
        self.fun = serial(fun) if (workers is None or workers <= 1) else parallel(fun, workers)  
        self.f = _fitness(self.fun, lower, upper, normalize)       
        if options is None:
            options = {}
        if not lower is None:
            options['constraint'] = [ [lower[i], upper[i]] for i in range(dim)]   
        self.constraint = options.get('constraint', [[-np.inf, np.inf] for _ in range(dim)])
        if 'seed' in options.keys():
            np.random.seed(options['seed'])
        sigma = input_sigma
        if not np.isscalar(sigma):
            sigma = np.mean(sigma)         
        self.m = np.array([self.f.encode(guess)]).T

        self.dim = dim
        self.sigma = sigma
        self.popsize = popsize
              
        self.max_evaluations = max_evaluations
        self.stop_fitness = stop_fitness
        self.is_terminate = is_terminate
        self.rg = rg
        self.runid = runid

        self.v = options.get('v', self.rg.normal(0,1,(dim, 1)) / np.sqrt(dim))
        
        self.D = np.ones([dim, 1])
        self.penalty_coef = options.get('penalty_coef', 1e5)
        self.use_constraint_violation = options.get('use_constraint_violation', True)

        self.w_rank_hat = (np.log(self.popsize / 2 + 1) - np.log(np.arange(1, self.popsize + 1))).reshape(self.popsize, 1)
        self.w_rank_hat[np.where(self.w_rank_hat < 0)] = 0
        self.w_rank = self.w_rank_hat / sum(self.w_rank_hat) - (1. / self.popsize)
        self.mueff = 1 / ((self.w_rank + (1 / self.popsize)).T @ (self.w_rank + (1 / self.popsize)))[0][0]
        self.cs = (self.mueff + 2.) / (self.dim + self.mueff + 5.)
        self.cc = (4. + self.mueff / self.dim) / (self.dim + 4. + 2. * self.mueff / self.dim)
        self.c1_cma = 2. / (math.pow(self.dim + 1.3, 2) + self.mueff)
        # initialization
        self.chiN = np.sqrt(self.dim) * (1. - 1. / (4. * self.dim) + 1. / (21. * self.dim * self.dim))
        self.pc = np.zeros([self.dim, 1])
        self.ps = np.zeros([self.dim, 1])
        # distance weight parameter
        self.h_inv = get_h_inv(self.dim)
        self.alpha_dist = lambda lambF: self.h_inv * min(1., math.sqrt(self.popsize / self.dim)) * math.sqrt(
            lambF / self.popsize)
        self.w_dist_hat = lambda z, lambF: exp(self.alpha_dist(lambF) * np.linalg.norm(z))
        # learning rate
        self.eta_m = 1.0
        self.eta_move_sigma = 1.
        self.eta_stag_sigma = lambda lambF: math.tanh((0.024 * lambF + 0.7 * self.dim + 20.) / (self.dim + 12.))
        self.eta_conv_sigma = lambda lambF: 2. * math.tanh((0.025 * lambF + 0.75 * self.dim + 10.) / (self.dim + 4.))
        self.c1 = lambda lambF: self.c1_cma * (self.dim - 5) / 6 * (lambF / self.popsize)
        self.eta_B = lambda lambF: np.tanh((min(0.02 * lambF, 3 * np.log(self.dim)) + 5) / (0.23 * self.dim + 25))

        self.g = 0
        self.no_of_evals = 0
        self.iteration = 0
        self.stop = 0

        self.idxp = np.arange(self.popsize / 2, dtype=int)
        self.idxm = np.arange(self.popsize / 2, self.popsize, dtype=int)
        self.z = np.zeros([self.dim, self.popsize])

        self.f_best = float('inf')
        self.x_best = np.empty(self.dim)

    def __del__(self):
        if isinstance(self.fun, parallel):
            self.fun.stop()
        
    def calc_violations(self, x):
        violations = np.zeros(self.popsize)
        for i in range(self.popsize):
            for j in range(self.dim):
                violations[i] += (- min(0, x[j][i] - self.constraint[j][0]) + max(0, x[j][i] - self.constraint[j][1])) * self.penalty_coef
        return violations

    def optimize(self) -> int:
        # -------------------- Generation Loop --------------------------------
        while True:
            if self.no_of_evals > self.max_evaluations:
                break
            if self.stop != 0:
                break
            try:
                x = self.ask()
                y = self.f.values(self.f.decode(self.f.closestFeasible(x)))
                self.tell(y)
                if self.stop != 0:
                    break 
            except Exception as ex:
                self.stop = -1
                break

    def ask(self) -> np.ndarray:
        d = self.dim
        popsize = self.popsize
        zhalf = self.rg.normal(0,1,(d, int(popsize / 2)))  # dim x popsize/2
        self.z[:, self.idxp] = zhalf
        self.z[:, self.idxm] = -zhalf
        self.normv = np.linalg.norm(self.v)
        self.normv2 = self.normv ** 2
        self.vbar = self.v / self.normv
        self.y = self.z + ((np.sqrt(1 + self.normv2) - 1) * (self.vbar @ (self.vbar.T @ self.z)))
        self.x = self.m + (self.sigma * self.y) * self.D
        return self.x.T

    def tell(self, evals_no_sort: np.ndarray) -> int:
        violations = np.zeros(self.popsize)
        if self.use_constraint_violation:
            violations = self.calc_violations(self.x)
            sorted_indices = sort_indices_by(evals_no_sort + violations, self.z)
        else:
            sorted_indices = sort_indices_by(evals_no_sort, self.z)
        best_eval_id = sorted_indices[0]
        f_best = evals_no_sort[best_eval_id]
        x_best = self.x[:, best_eval_id]
        self.z = self.z[:, sorted_indices]
        y = self.y[:, sorted_indices]
        x = self.x[:, sorted_indices]

        self.no_of_evals += self.popsize
        self.g += 1
 
        if f_best < self.f_best:
            self.f_best = f_best
            self.x_best = x_best           
            # print(self.no_of_evals, self.g, self.f_best)

        # This operation assumes that if the solution is infeasible, infinity comes in as input.
        lambF = np.sum(evals_no_sort < np.finfo(float).max)

        # evolution path p_sigma
        self.ps = (1 - self.cs) * self.ps + np.sqrt(self.cs * (2. - self.cs) * self.mueff) * (self.z @ self.w_rank)
        ps_norm = np.linalg.norm(self.ps)
        # distance weight
        f1 =  self.h_inv * min(1., math.sqrt(self.popsize / self.dim)) * math.sqrt(lambF / self.popsize)        
        w_tmp = self.w_rank_hat * np.exp(np.linalg.norm(self.z, axis = 0) * f1).reshape((self.popsize,1))
        weights_dist = w_tmp / sum(w_tmp) - 1. / self.popsize
        # switching weights and learning rate
        weights = weights_dist if ps_norm >= self.chiN else self.w_rank
        eta_sigma = self.eta_move_sigma if ps_norm >= self.chiN else self.eta_stag_sigma(
            lambF) if ps_norm >= 0.1 * self.chiN else self.eta_conv_sigma(lambF)
        # update pc, m
        wxm = (x - self.m) @ weights
        self.pc = (1. - self.cc) * self.pc + np.sqrt(self.cc * (2. - self.cc) * self.mueff) * wxm / self.sigma
        self.m += self.eta_m * wxm
        # calculate s, t
        # step1
        normv4 = self.normv2 ** 2
        exY = np.append(y, self.pc / self.D, axis=1)  # dim x popsize+1
        yy = exY * exY  # dim x popsize+1
        ip_yvbar = self.vbar.T @ exY
        yvbar = exY * self.vbar  # dim x popsize+1. exYのそれぞれの列にvbarがかかる
        gammav = 1. + self.normv2
        vbarbar = self.vbar * self.vbar
        alphavd = min(1, np.sqrt(normv4 + (2 * gammav - np.sqrt(gammav)) / np.max(vbarbar)) / (2 + self.normv2))  # scalar
        
        t = exY * ip_yvbar - self.vbar * (ip_yvbar ** 2 + gammav) / 2  # dim x popsize+1
        b = -(1 - alphavd ** 2) * normv4 / gammav + 2 * alphavd ** 2
        H = np.ones([self.dim, 1]) * 2 - (b + 2 * alphavd ** 2) * vbarbar  # dim x 1
        invH = H ** (-1)
        s_step1 = yy - self.normv2 / gammav * (yvbar * ip_yvbar) - np.ones([self.dim, self.popsize + 1])  # dim x popsize+1
        ip_vbart = self.vbar.T @ t  # 1 x popsize+1
 
        s_step2 = s_step1 - alphavd / gammav * ((2 + self.normv2) * (t * self.vbar) - self.normv2 * vbarbar @ ip_vbart)  # dim x popsize+1
        invHvbarbar = invH * vbarbar
        ip_s_step2invHvbarbar = invHvbarbar.T @ s_step2  # 1 x popsize+1
        
        div = 1 + b * vbarbar.T @ invHvbarbar
        if np.amin(abs(div)) == 0:
            return -1
        
        s = (s_step2 * invH) - b / div * invHvbarbar @ ip_s_step2invHvbarbar  # dim x popsize+1
        ip_svbarbar = vbarbar.T @ s  # 1 x popsize+1
        t = t - alphavd * ((2 + self.normv2) * (s * self.vbar) - self.vbar @ ip_svbarbar)  # dim x popsize+1
        # update v, D
        exw = np.append(self.eta_B(lambF) * weights, np.array([self.c1(lambF)]).reshape(1, 1),
                        axis=0)  # popsize+1 x 1
        self.v = self.v + (t @ exw) / self.normv
        self.D = self.D + (s @ exw) * self.D
        # calculate detA
        if np.amin(self.D) < 0:
            return -1

        nthrootdetA = exp(np.sum(np.log(self.D)) / self.dim + np.log(1 + (self.v.T @ self.v)[0][0]) / (2 * self.dim))
         
        self.D = self.D / nthrootdetA
        
        # update sigma
        G_s = np.sum((self.z * self.z - np.ones([self.dim, self.popsize])) @ weights) / self.dim
        self.sigma = self.sigma * exp(eta_sigma / 2 * G_s)
        return self.stop

    def population(self) -> np.ndarray:
        return self.x

    def result(self) -> OptimizeResult:
        return OptimizeResult(x=self.x_best, fun=self.f_best, nfev=self.no_of_evals, 
                              nit=self.g, status=self.stop, success=True)
        
def exp(a):
    return math.exp(min(100, a)) # avoid overflow

def get_h_inv(dim):
    f = lambda a, b: ((1. + a * a) * exp(a * a / 2.) / 0.24) - 10. - dim
    f_prime = lambda a: (1. / 0.24) * a * exp(a * a / 2.) * (3. + a * a)
    h_inv = 1.0
    while abs(f(h_inv, dim)) > 1e-10:
        h_inv = h_inv - 0.5 * (f(h_inv, dim) / f_prime(h_inv))
    return h_inv

def sort_indices_by(evals, z):
    lam = len(evals)
    evals = np.array(evals)
    sorted_indices = np.argsort(evals)
    sorted_evals = evals[sorted_indices]
    no_of_feasible_solutions = np.where(sorted_evals != INFEASIBLE)[0].size
    if no_of_feasible_solutions != lam:
        infeasible_z = z[:, np.where(evals == INFEASIBLE)[0]]
        distances = np.sum(infeasible_z ** 2, axis=0)
        infeasible_indices = sorted_indices[no_of_feasible_solutions:]
        indices_sorted_by_distance = np.argsort(distances)
        sorted_indices[no_of_feasible_solutions:] = infeasible_indices[indices_sorted_by_distance]
    return sorted_indices



# Copyright (c) Dietmar Wolz.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory.

""" Eigen based implementation of active CMA-ES.
    Derived from http://cma.gforge.inria.fr/cmaes.m which follows
    https://www.researchgate.net/publication/227050324_The_CMA_Evolution_Strategy_A_Comparing_Review
"""

import sys
import os
import math
import ctypes as ct
import numpy as np
from numpy.random import MT19937, Generator
from scipy.optimize import OptimizeResult, Bounds
from fcmaes.evaluator import _check_bounds, _get_bounds, mo_call_back_type, callback_so, callback_par, call_back_par, parallel, libcmalib

from typing import Optional, Callable, Union
from numpy.typing import ArrayLike

os.environ['MKL_DEBUG_CPU_TYPE'] = '5'

def minimize(fun: Callable[[ArrayLike], float], 
             bounds: Optional[Bounds] = None, 
             x0: Optional[ArrayLike] = None,
             input_sigma: Optional[Union[float, ArrayLike, Callable]] = 0.3,
             popsize: Optional[int] = 31,
             max_evaluations: Optional[int]  = 100000,
             accuracy: Optional[float] = 1.0, 
             stop_fitness: Optional[float] = -np.inf, 
             stop_hist: Optional[float] = None,
             rg: Optional[Generator] = Generator(MT19937()),
             runid: Optional[int] = 0,
             workers: Optional[int] = 1, 
             normalize: Optional[bool] = True,
             delayed_update: Optional[bool] = True,
             update_gap: Optional[int] = None
             ) -> OptimizeResult:
   
    """Minimization of a scalar function of one or more variables using a 
    C++ CMA-ES implementation called via ctypes.
     
    Parameters
    ----------
    fun : callable
        The objective function to be minimized.
            ``fun(x) -> float``
        where ``x`` is an 1-D array with shape (dim,)
    bounds : sequence or `Bounds`, optional
        Bounds on variables. There are two ways to specify the bounds:
            1. Instance of the `scipy.Bounds` class.
            2. Sequence of ``(min, max)`` pairs for each element in `x`. None
               is used to specify no bound.
    x0 : ndarray, shape (dim,)
        Initial guess. Array of real elements of size (dim,),
        where 'dim' is the number of independent variables.  
    input_sigma : ndarray, shape (dim,) or scalar
        Initial step size for each dimension.
    popsize = int, optional
        CMA-ES population size.
    max_evaluations : int, optional
        Forced termination after ``max_evaluations`` function evaluations.
    accuracy : float, optional
        values > 1.0 reduce the accuracy.
    stop_fitness : float, optional 
         Limit for fitness value. If reached minimize terminates.
    stop_hist : float, optional 
         Set to 0 if you want to prevent premature termination because 
         there is no progress
    rg = numpy.random.Generator, optional
        Random generator for creating random guesses.
    runid : int, optional
        id used by the is_terminate callback to identify the CMA-ES run. 
    workers : int or None, optional
        If not workers is None, function evaluation is performed in parallel for the whole population. 
        Useful for costly objective functions but is deactivated for parallel retry.      
    normalize : boolean, optional
        if true pheno -> geno transformation maps arguments to interval [-1,1]
    delayed_update : boolean, optional
        if true uses delayed update / C++ parallelism, i false uses Python multithreading
    update_gap : int, optional
        number of iterations without distribution update
           
    Returns
    -------
    res : scipy.OptimizeResult
        The optimization result is represented as an ``OptimizeResult`` object.
        Important attributes are: ``x`` the solution array, 
        ``fun`` the best function value, 
        ``nfev`` the number of function evaluations,
        ``nit`` the number of CMA-ES iterations, 
        ``status`` the stopping critera and
        ``success`` a Boolean flag indicating if the optimizer exited successfully. """
    
    lower, upper, guess = _check_bounds(bounds, x0, rg)      
    dim = guess.size   
    if lower is None:
        lower = [0]*dim
        upper = [0]*dim
    if workers is None:
        workers = 0
    mu = int(popsize/2)
    if callable(input_sigma):
        input_sigma=input_sigma()
    if np.ndim(input_sigma) == 0:
        input_sigma = [input_sigma] * dim
    if stop_hist is None:
        stop_hist = -1;
    array_type = ct.c_double * dim 
    c_callback = mo_call_back_type(callback_so(fun, dim))
    parfun = None if delayed_update == True or workers is None or workers <= 1 else parallel(fun, workers)
    c_callback_par = call_back_par(callback_par(fun, parfun))
    res = np.empty(dim+4)
    res_p = res.ctypes.data_as(ct.POINTER(ct.c_double))
    try:
        optimizeACMA_C(runid, c_callback, c_callback_par, 
                dim, array_type(*guess), array_type(*lower), array_type(*upper), 
                array_type(*input_sigma), max_evaluations, stop_fitness, stop_hist, mu, 
                popsize, accuracy, int(rg.uniform(0, 2**32 - 1)), 
                normalize, delayed_update, -1 if update_gap is None else update_gap,
                workers, res_p)
        x = res[:dim]
        val = res[dim]
        evals = int(res[dim+1])
        iterations = int(res[dim+2])
        stop = int(res[dim+3])
        res = OptimizeResult(x=x, fun=val, nfev=evals, nit=iterations, status=stop, success=True)
    except Exception as ex:
        res = OptimizeResult(x=None, fun=sys.float_info.max, nfev=0, nit=0, status=-1, success=False)
    if not parfun is None:
        parfun.stop()
    return res

class ACMA_C:

    def __init__(self,
        dim, 
        bounds: Optional[Bounds] = None, 
        x0: Optional[ArrayLike] = None,
        input_sigma: Optional[Union[float, ArrayLike, Callable]] = 0.3, 
        popsize: Optional[int] = 31,  
        max_evaluations: Optional[int] = 100000, 
        accuracy: Optional[float] = 1.0, 
        stop_fitness: Optional[float] = -np.inf, 
        stop_hist: Optional[float] = None,
        rg: Optional[Generator] = Generator(MT19937()),
        runid: Optional[int] = 0,
        normalize: Optional[bool] = True,
        delayed_update: Optional[bool] = True,
        update_gap: Optional[int] = None
     ):
       
        """Parameters
        ----------
        dim : int
            dimension of the argument of the objective function
        bounds : sequence or `Bounds`, optional
            Bounds on variables. There are two ways to specify the bounds:
                1. Instance of the `scipy.Bounds` class.
                2. Sequence of ``(min, max)`` pairs for each element in `x`. None
                   is used to specify no bound.
        x0 : ndarray, shape (dim,)
            Initial guess. Array of real elements of size (dim,),
            where 'dim' is the number of independent variables.  
        input_sigma : ndarray, shape (dim,) or scalar
            Initial step size for each dimension.
        popsize = int, optional
            CMA-ES population size.
        max_evaluations : int, optional
            Forced termination after ``max_evaluations`` function evaluations.
        accuracy : float, optional
            values > 1.0 reduce the accuracy.
        stop_fitness : float, optional 
             Limit for fitness value. If reached minimize terminates.
        stop_hist : float, optional 
             Set to 0 if you want to prevent premature termination because 
             there is no progress
        rg = numpy.random.Generator, optional
            Random generator for creating random guesses.
        runid : int, optional
            id used by the is_terminate callback to identify the CMA-ES run.     
        normalize : boolean, optional
            if true pheno -> geno transformation maps arguments to interval [-1,1]
        delayed_update : boolean, optional
            if true uses delayed update / C++ parallelism, i false uses Python multithreading
        update_gap : int, optional
            number of iterations without distribution update"""
             
        lower, upper, guess = _get_bounds(dim, bounds, x0, rg)     
        if lower is None:
            lower = [0]*dim
            upper = [0]*dim
        mu = int(popsize/2)
        if callable(input_sigma):
            input_sigma=input_sigma()
        if np.ndim(input_sigma) == 0:
            input_sigma = [input_sigma] * dim
        if stop_hist is None:
            stop_hist = -1;
        array_type = ct.c_double * dim 
        try:
            self.ptr = initACMA_C(runid,
                dim, array_type(*guess), array_type(*lower), array_type(*upper), 
                array_type(*input_sigma), max_evaluations, stop_fitness, stop_hist, mu, 
                popsize, accuracy, int(rg.uniform(0, 2**32 - 1)), 
                normalize, delayed_update, -1 if update_gap is None else update_gap)
            self.popsize = popsize
            self.dim = dim            
        except Exception as ex:
            print (ex)
            pass
    
    def __del__(self):
        destroyACMA_C(self.ptr)
            
    def ask(self) -> np.array:
        try:
            popsize = self.popsize
            n = self.dim
            res = np.empty(popsize*n)
            res_p = res.ctypes.data_as(ct.POINTER(ct.c_double))
            askACMA_C(self.ptr, res_p)
            xs = np.empty((popsize, n))
            for p in range(popsize):
                xs[p,:] = res[p*n : (p+1)*n]
            return xs
        except Exception as ex:
            print (ex)
            return None

    def tell(self, 
             ys: np.ndarray, 
             xs: Optional[np.ndarray] = None) -> int:  
        if not xs is None:
            return self.tell_x_(ys, xs)
        try:
            array_type_ys = ct.c_double * len(ys)
            return tellACMA_C(self.ptr, array_type_ys(*ys))
        except Exception as ex:
            print (ex)
            return -1    

    def tell_x_(self, ys: np.ndarray, xs: np.ndarray):
        try:
            flat_xs = xs.flatten()
            array_type_xs = ct.c_double * len(flat_xs)
            array_type_ys = ct.c_double * len(ys)
            return tellXACMA_C(self.ptr, array_type_ys(*ys), array_type_xs(*flat_xs))
        except Exception as ex:
            print (ex)
            return -1 
        
    def population(self) -> np.array:
        try:
            popsize = self.popsize
            n = self.dim
            res = np.empty(popsize*n)
            res_p = res.ctypes.data_as(ct.POINTER(ct.c_double))
            populationACMA_C(self.ptr, res_p)
            xs = np.array(popsize, n)
            for p in range(popsize):
                xs[p] = res[p*n : (p+1)*n]
                return xs
        except Exception as ex:
            print (ex)
            return None

    def result(self) -> OptimizeResult:
        res = np.empty(self.dim+4)
        res_p = res.ctypes.data_as(ct.POINTER(ct.c_double))
        try:
            resultACMA_C(self.ptr, res_p)
            x = res[:self.dim]
            val = res[self.dim]
            evals = int(res[self.dim+1])
            iterations = int(res[self.dim+2])
            stop = int(res[self.dim+3])
            res = OptimizeResult(x=x, fun=val, nfev=evals, nit=iterations, status=stop, success=True)
        except Exception as ex:
            res = OptimizeResult(x=None, fun=sys.float_info.max, nfev=0, nit=0, status=-1, success=False)
        return res

if not libcmalib is None: 

    optimizeACMA_C = libcmalib.optimizeACMA_C
    optimizeACMA_C.argtypes = [ct.c_long, mo_call_back_type, call_back_par, ct.c_int, \
                ct.POINTER(ct.c_double), ct.POINTER(ct.c_double), ct.POINTER(ct.c_double), \
                ct.POINTER(ct.c_double), ct.c_int, ct.c_double, ct.c_double, ct.c_int, ct.c_int, \
                ct.c_double, ct.c_long, ct.c_bool, ct.c_bool, ct.c_int, 
                ct.c_int, ct.POINTER(ct.c_double)]
    
    initACMA_C = libcmalib.initACMA_C
    initACMA_C.argtypes = [ct.c_long, ct.c_int, \
                ct.POINTER(ct.c_double), ct.POINTER(ct.c_double), ct.POINTER(ct.c_double), \
                ct.POINTER(ct.c_double), ct.c_int, ct.c_double, ct.c_double, ct.c_int, 
                ct.c_int, ct.c_double, ct.c_long, ct.c_bool, ct.c_bool, ct.c_int]
                    
    initACMA_C.restype = ct.c_void_p   
    
    destroyACMA_C = libcmalib.destroyACMA_C
    destroyACMA_C.argtypes = [ct.c_void_p]
    
    askACMA_C = libcmalib.askACMA_C
    askACMA_C.argtypes = [ct.c_void_p, ct.POINTER(ct.c_double)]
    
    tellACMA_C = libcmalib.tellACMA_C
    tellACMA_C.argtypes = [ct.c_void_p, ct.POINTER(ct.c_double)]
    tellACMA_C.restype = ct.c_int
    
    tellXACMA_C = libcmalib.tellXACMA_C
    tellXACMA_C.argtypes = [ct.c_void_p, ct.POINTER(ct.c_double), ct.POINTER(ct.c_double)]
    tellXACMA_C.restype = ct.c_int
    
    populationACMA_C = libcmalib.populationACMA_C
    populationACMA_C.argtypes = [ct.c_void_p, ct.POINTER(ct.c_double)]
    
    resultACMA_C = libcmalib.resultACMA_C
    resultACMA_C.argtypes = [ct.c_void_p, ct.POINTER(ct.c_double)]
